# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
import pyomo.common.unittest as unittest

from pyomo.environ import (
    ConcreteModel,
    Block,
    Var,
    Reference,
    Set,
    Constraint,
    ComponentUID,
)
from pyomo.dae import ContinuousSet
from pyomo.common.collections import ComponentSet, ComponentMap
from pyomo.core.base.indexed_component import UnindexedComponent_set, normalize_index
from pyomo.dae.flatten import (
    flatten_dae_components,
    flatten_components_along_sets,
    slice_component_along_sets,
)


class _TestFlattenBase:
    """A base class to hold the common _hashRef utility method.
    We don't just derive from Test... classes directly as this
    causes tests to run twice.

    """

    def _hashRef(self, ref):
        if not ref.is_indexed():
            return (id(ref),)
        else:
            return tuple(sorted(id(_) for _ in ref.values()))


class TestAssumedBehavior(unittest.TestCase):
    """
    These are some behaviors we rely on that weren't
    immediately obvious would be the case.
    """

    def setUp(self):
        self._orig_flatten = normalize_index.flatten

    def tearDown(self):
        normalize_index.flatten = self._orig_flatten

    def test_cross(self):
        m = ConcreteModel()
        m.s1 = Set(initialize=[1, 2])
        m.s2 = Set(initialize=[3, 4])
        m.s3 = Set(initialize=['a', 'b'])

        normalize_index.flatten = True

        for i in m.s1.cross():
            # A "vacuous cross product" will place set elements in tuples
            self.assertIs(type(i), tuple)

        for i in m.s1.cross(m.s2, m.s3):
            self.assertIs(type(i), tuple)
            for j in i:
                # A cross product with multiple arguments does not produce
                # nested tuples
                self.assertIsNot(type(j), tuple)

        normalize_index.flatten = False
        # This behavior is consistent regardless of the value of
        # normalize_index.flatten

        for i in m.s1.cross():
            self.assertIs(type(i), tuple)
        for i in m.s1.cross(m.s2, m.s3):
            self.assertIs(type(i), tuple)
            for j in i:
                self.assertIsNot(type(j), tuple)

        normalize_index.flatten = True

    def test_subsets(self):
        m = ConcreteModel()
        m.s1 = Set(initialize=[1, 2])
        m.s2 = Set(initialize=[3, 4])
        m.s3 = Set(initialize=['a', 'b'])

        normalize_index.flatten = True

        s12 = m.s1 * m.s2
        s12_3 = s12 * m.s3

        s123 = m.s1.cross(m.s2, m.s3)

        subsets12_3 = list(s12_3.subsets())
        subsets123 = list(s123.subsets())

        # `subsets` identifies the "base sets" regardless
        # of whether products are nested.

        self.assertEqual(len(subsets12_3), len(subsets123))
        for s_a, s_b in zip(subsets12_3, subsets123):
            self.assertIs(s_a, s_b)

        normalize_index.flatten = False

        for i, j in s12_3:
            # Make sure we have a nested product with flatten False.
            self.assertIs(type(i), tuple)

        # The behavior of subsets is unchanged
        subsets12_3 = list(s12_3.subsets())
        subsets123 = list(s123.subsets())

        self.assertEqual(len(subsets12_3), len(subsets123))
        for s_a, s_b in zip(subsets12_3, subsets123):
            self.assertIs(s_a, s_b)

        normalize_index.flatten = True


class TestCategorize(_TestFlattenBase, unittest.TestCase):
    def test_flat_model(self):
        m = ConcreteModel()
        m.T = ContinuousSet(bounds=(0, 1))
        m.x = Var()
        m.y = Var([1, 2])
        m.a = Var(m.T)
        m.b = Var(m.T, [1, 2])
        m.c = Var([3, 4], m.T)

        regular, time = flatten_dae_components(m, m.T, Var)
        regular_id = set(id(_) for _ in regular)
        self.assertEqual(len(regular), 3)
        self.assertIn(id(m.x), regular_id)
        self.assertIn(id(m.y[1]), regular_id)
        self.assertIn(id(m.y[2]), regular_id)
        # Output for debugging
        # for v in time:
        #    v.pprint()
        #    for _ in v.values():
        #        print"     -> ", _.name
        ref_data = {
            self._hashRef(Reference(m.a[:])),
            self._hashRef(Reference(m.b[:, 1])),
            self._hashRef(Reference(m.b[:, 2])),
            self._hashRef(Reference(m.c[3, :])),
            self._hashRef(Reference(m.c[4, :])),
        }
        self.assertEqual(len(time), len(ref_data))
        for ref in time:
            self.assertIn(self._hashRef(ref), ref_data)

    def test_1level_model(self):
        m = ConcreteModel()
        m.T = ContinuousSet(bounds=(0, 1))

        @m.Block([1, 2], m.T)
        def B(b, i, t):
            b.x = Var(list(range(2 * i, 2 * i + 2)))

        regular, time = flatten_dae_components(m, m.T, Var)
        self.assertEqual(len(regular), 0)
        # Output for debugging
        # for v in time:
        #    v.pprint()
        #    for _ in v.values():
        #        print"     -> ", _.name
        ref_data = {
            self._hashRef(Reference(m.B[1, :].x[2])),
            self._hashRef(Reference(m.B[1, :].x[3])),
            self._hashRef(Reference(m.B[2, :].x[4])),
            self._hashRef(Reference(m.B[2, :].x[5])),
        }
        self.assertEqual(len(time), len(ref_data))
        for ref in time:
            self.assertIn(self._hashRef(ref), ref_data)

    def test_2level_model(self):
        m = ConcreteModel()
        m.T = ContinuousSet(bounds=(0, 1))

        @m.Block([1, 2], m.T)
        def B(b, i, t):
            @b.Block(list(range(2 * i, 2 * i + 2)))
            def bb(bb, j):
                bb.y = Var([10, 11])

            b.x = Var(list(range(2 * i, 2 * i + 2)))

        regular, time = flatten_dae_components(m, m.T, Var)
        self.assertEqual(len(regular), 0)
        # Output for debugging
        # for v in time:
        #    v.pprint()
        #    for _ in v.values():
        #        print"     -> ", _.name
        ref_data = {
            self._hashRef(Reference(m.B[1, :].x[2])),
            self._hashRef(Reference(m.B[1, :].x[3])),
            self._hashRef(Reference(m.B[2, :].x[4])),
            self._hashRef(Reference(m.B[2, :].x[5])),
            self._hashRef(Reference(m.B[1, :].bb[2].y[10])),
            self._hashRef(Reference(m.B[1, :].bb[2].y[11])),
            self._hashRef(Reference(m.B[1, :].bb[3].y[10])),
            self._hashRef(Reference(m.B[1, :].bb[3].y[11])),
            self._hashRef(Reference(m.B[2, :].bb[4].y[10])),
            self._hashRef(Reference(m.B[2, :].bb[4].y[11])),
            self._hashRef(Reference(m.B[2, :].bb[5].y[10])),
            self._hashRef(Reference(m.B[2, :].bb[5].y[11])),
        }
        self.assertEqual(len(time), len(ref_data))
        for ref in time:
            self.assertIn(self._hashRef(ref), ref_data)

    def test_2dim_set(self):
        m = ConcreteModel()
        m.time = ContinuousSet(bounds=(0, 1))

        m.v = Var(m.time, [('a', 1), ('b', 2)])

        scalar, dae = flatten_dae_components(m, m.time, Var)
        self.assertEqual(len(scalar), 0)
        ref_data = {
            self._hashRef(Reference(m.v[:, 'a', 1])),
            self._hashRef(Reference(m.v[:, 'b', 2])),
        }
        self.assertEqual(len(dae), len(ref_data))
        for ref in dae:
            self.assertIn(self._hashRef(ref), ref_data)

    def test_indexed_block(self):
        m = ConcreteModel()
        m.time = ContinuousSet(bounds=(0, 1))
        m.comp = Set(initialize=['a', 'b'])

        def bb_rule(bb, t):
            bb.dae_var = Var()

        def b_rule(b, c):
            b.bb = Block(m.time, rule=bb_rule)

        m.b = Block(m.comp, rule=b_rule)

        scalar, dae = flatten_dae_components(m, m.time, Var)
        self.assertEqual(len(scalar), 0)
        ref_data = {
            self._hashRef(Reference(m.b['a'].bb[:].dae_var)),
            self._hashRef(Reference(m.b['b'].bb[:].dae_var)),
        }
        self.assertEqual(len(dae), len(ref_data))
        for ref in dae:
            self.assertIn(self._hashRef(ref), ref_data)

    def test_constraint(self):
        m = ConcreteModel()
        m.time = ContinuousSet(bounds=(0, 1))
        m.comp = Set(initialize=['a', 'b'])
        m.v0 = Var()
        m.v1 = Var(m.time)
        m.v2 = Var(m.time, m.comp)

        def c0_rule(m):
            return m.v0 == 1

        m.c0 = Constraint(rule=c0_rule)

        def c1_rule(m, t):
            return m.v1[t] == 3

        m.c1 = Constraint(m.time, rule=c1_rule)

        @m.Block(m.time)
        def b(b, t):
            def c2_rule(b, j):
                return b.model().v2[t, j] == 5

            b.c2 = Constraint(m.comp, rule=c2_rule)

        scalar, dae = flatten_dae_components(m, m.time, Constraint)
        hash_scalar = {id(s) for s in scalar}
        self.assertIn(id(m.c0), hash_scalar)

        ref_data = {
            self._hashRef(Reference(m.c1[:])),
            self._hashRef(Reference(m.b[:].c2['a'])),
            self._hashRef(Reference(m.b[:].c2['b'])),
        }
        self.assertEqual(len(dae), len(ref_data))
        for ref in dae:
            self.assertIn(self._hashRef(ref), ref_data)

    def test_constraint_skip(self):
        m = ConcreteModel()
        m.time = ContinuousSet(bounds=(0, 1))

        m.v = Var(m.time)

        def c_rule(m, t):
            if t == m.time.first():
                return Constraint.Skip
            return m.v[t] == 1.0

        m.c = Constraint(m.time, rule=c_rule)
        scalar, dae = flatten_dae_components(m, m.time, Constraint)

        ref_data = {self._hashRef(Reference(m.c[:]))}
        self.assertEqual(len(dae), len(ref_data))
        for ref in dae:
            self.assertIn(self._hashRef(ref), ref_data)


class TestFlatten(_TestFlattenBase, unittest.TestCase):
    def setUp(self):
        self._orig_flatten = normalize_index.flatten

    def tearDown(self):
        normalize_index.flatten = self._orig_flatten

    def _model1_1d_sets(self):
        # One-dimensional sets, no skipping.
        m = ConcreteModel()
        m.time = Set(initialize=[1, 2, 3])
        m.space = Set(initialize=[0.0, 0.5, 1.0])
        m.comp = Set(initialize=['a', 'b'])

        m.v0 = Var()
        m.v1 = Var(m.time)
        m.v2 = Var(m.time, m.space)
        m.v3 = Var(m.time, m.space, m.comp)

        m.v_tt = Var(m.time, m.time)
        m.v_tst = Var(m.time, m.space, m.time)

        @m.Block()
        def b(b):
            @b.Block(m.time)
            def b1(b1):
                b1.v0 = Var()
                b1.v1 = Var(m.space)
                b1.v2 = Var(m.space, m.comp)

                @b1.Block(m.space)
                def b_s(b_s):
                    b_s.v0 = Var()
                    b_s.v1 = Var(m.space)
                    b_s.v2 = Var(m.space, m.comp)

            @b.Block(m.time, m.space)
            def b2(b2):
                b2.v0 = Var()
                b2.v1 = Var(m.comp)
                b2.v2 = Var(m.time, m.comp)

        return m

    def test_flatten_m1_along_time_space(self):
        m = self._model1_1d_sets()

        sets = ComponentSet((m.time, m.space))
        sets_list, comps_list = flatten_components_along_sets(m, sets, Var)
        assert len(sets_list) == len(comps_list)
        assert len(sets_list) == 6

        for sets, comps in zip(sets_list, comps_list):
            if len(sets) == 1 and sets[0] is UnindexedComponent_set:
                ref_data = {self._hashRef(m.v0)}
                assert len(comps) == len(ref_data)
                for comp in comps:
                    self.assertIn(self._hashRef(comp), ref_data)

            elif len(sets) == 1 and sets[0] is m.time:
                ref_data = {
                    self._hashRef(Reference(m.v1)),
                    self._hashRef(Reference(m.b.b1[:].v0)),
                }
                assert len(comps) == len(ref_data)
                for comp in comps:
                    self.assertIn(self._hashRef(comp), ref_data)

            elif len(sets) == 2 and sets[0] is m.time and sets[1] is m.time:
                ref_data = {self._hashRef(Reference(m.v_tt))}
                assert len(comps) == len(ref_data)
                for comp in comps:
                    self.assertIn(self._hashRef(comp), ref_data)

            elif len(sets) == 2 and sets[0] is m.time and sets[1] is m.space:
                ref_data = {
                    self._hashRef(m.v2),
                    self._hashRef(Reference(m.b.b1[:].v1[:])),
                    self._hashRef(Reference(m.b.b2[:, :].v0)),
                    self._hashRef(Reference(m.b.b1[:].b_s[:].v0)),
                }
                ref_data.update(self._hashRef(Reference(m.v3[:, :, j])) for j in m.comp)
                ref_data.update(
                    self._hashRef(Reference(m.b.b1[:].v2[:, j])) for j in m.comp
                )
                ref_data.update(
                    self._hashRef(Reference(m.b.b2[:, :].v1[j])) for j in m.comp
                )

                assert len(comps) == len(ref_data)
                for comp in comps:
                    self.assertIn(self._hashRef(comp), ref_data)

            elif (
                len(sets) == 3
                and sets[0] is m.time
                and sets[1] is m.space
                and sets[2] is m.time
            ):
                ref_data = {self._hashRef(m.v_tst)}
                ref_data.update(
                    self._hashRef(Reference(m.b.b2[:, :].v2[:, j])) for j in m.comp
                )
                assert len(comps) == len(ref_data)
                for comp in comps:
                    self.assertIn(self._hashRef(comp), ref_data)

            elif (
                len(sets) == 3
                and sets[0] is m.time
                and sets[1] is m.space
                and sets[2] is m.space
            ):
                ref_data = {self._hashRef(Reference(m.b.b1[:].b_s[:].v1[:]))}
                ref_data.update(
                    self._hashRef(Reference(m.b.b1[:].b_s[:].v2[:, j])) for j in m.comp
                ),
                assert len(comps) == len(ref_data)
                for comp in comps:
                    self.assertIn(self._hashRef(comp), ref_data)

            else:
                raise RuntimeError()

    def test_flatten_m1_empty(self):
        m = self._model1_1d_sets()

        sets = ComponentSet()
        sets_list, comps_list = flatten_components_along_sets(m, sets, Var)
        assert len(sets_list) == len(comps_list)
        assert len(sets_list) == 1

        for sets, comps in zip(sets_list, comps_list):
            if len(sets) == 1 and sets[0] is UnindexedComponent_set:
                ref_data = {self._hashRef(v) for v in m.component_data_objects(Var)}
                assert len(comps) == len(ref_data)
                for comp in comps:
                    self.assertIn(self._hashRef(comp), ref_data)
            else:
                raise RuntimeError()

    def test_flatten_m1_along_space(self):
        m = self._model1_1d_sets()

        sets = ComponentSet((m.space,))
        sets_list, comps_list = flatten_components_along_sets(m, sets, Var)
        assert len(sets_list) == len(comps_list)
        assert len(sets_list) == 3

        T = m.time
        TC = m.time * m.comp
        TT = m.time * m.time
        TTC = m.time * m.time * m.comp
        # These products are nested, i.e. ((t,t),j). This is fine
        # as normalize_index.flatten is True.

        for sets, comps in zip(sets_list, comps_list):
            if len(sets) == 1 and sets[0] is UnindexedComponent_set:
                ref_data = {self._hashRef(m.v0)}
                ref_data.update(self._hashRef(m.v1[t]) for t in T)
                ref_data.update(self._hashRef(m.v_tt[t1, t2]) for t1, t2 in TT)
                ref_data.update(self._hashRef(m.b.b1[t].v0) for t in T)
                assert len(comps) == len(ref_data)
                for comp in comps:
                    self.assertIn(self._hashRef(comp), ref_data)
            elif len(sets) == 1 and sets[0] is m.space:
                ref_data = set()
                ref_data.update(self._hashRef(Reference(m.v2[t, :])) for t in T)
                ref_data.update(self._hashRef(Reference(m.v3[t, :, j])) for t, j in TC)
                ref_data.update(
                    self._hashRef(Reference(m.v_tst[t1, :, t2])) for t1, t2 in TT
                )
                ref_data.update(self._hashRef(Reference(m.b.b1[t].v1[:])) for t in T)
                ref_data.update(
                    self._hashRef(Reference(m.b.b1[t].v2[:, j])) for t, j in TC
                )
                ref_data.update(
                    self._hashRef(Reference(m.b.b1[t].b_s[:].v0)) for t in T
                )
                ref_data.update(self._hashRef(Reference(m.b.b2[t, :].v0)) for t in T)
                ref_data.update(
                    self._hashRef(Reference(m.b.b2[t, :].v1[j])) for t, j in TC
                )
                ref_data.update(
                    self._hashRef(Reference(m.b.b2[t1, :].v2[t2, j]))
                    for t1, t2, j in TTC
                )
                assert len(comps) == len(ref_data)
                for comp in comps:
                    self.assertIn(self._hashRef(comp), ref_data)

            elif len(sets) == 2 and sets[0] is m.space and sets[1] is m.space:
                ref_data = set()
                ref_data.update(
                    self._hashRef(Reference(m.b.b1[t].b_s[:].v1[:])) for t in T
                )
                ref_data.update(
                    self._hashRef(Reference(m.b.b1[t].b_s[:].v2[:, j])) for t, j in TC
                )
                assert len(comps) == len(ref_data)
                for comp in comps:
                    self.assertIn(self._hashRef(comp), ref_data)
            else:
                raise RuntimeError()

    def test_flatten_m1_along_time(self):
        m = self._model1_1d_sets()

        sets = ComponentSet((m.time,))
        sets_list, comps_list = flatten_components_along_sets(m, sets, Var)

        S = m.space
        SS = m.space * m.space
        SC = m.space * m.comp
        SSC = m.space * m.space * m.comp

        assert len(sets_list) == 3
        for sets, comps in zip(sets_list, comps_list):
            if len(sets) == 1 and sets[0] is UnindexedComponent_set:
                ref_data = {self._hashRef(Reference(m.v0))}
                assert len(comps) == len(ref_data)
                for comp in comps:
                    self.assertIn(self._hashRef(comp), ref_data)
            elif len(sets) == 1 and sets[0] is m.time:
                ref_data = {
                    # Components indexed only by time;
                    self._hashRef(Reference(m.v1)),
                    self._hashRef(Reference(m.b.b1[:].v0)),
                }
                # Components indexed by time and some other set(s)
                ref_data.update(self._hashRef(Reference(m.v2[:, x])) for x in S)
                ref_data.update(self._hashRef(Reference(m.v3[:, x, j])) for x, j in SC)
                ref_data.update(self._hashRef(Reference(m.b.b1[:].v1[x])) for x in S)
                ref_data.update(
                    self._hashRef(Reference(m.b.b1[:].v2[x, j])) for x, j in SC
                )
                ref_data.update(
                    self._hashRef(Reference(m.b.b1[:].b_s[x].v0)) for x in S
                )
                ref_data.update(
                    self._hashRef(Reference(m.b.b1[:].b_s[x1].v1[x2])) for x1, x2 in SS
                )
                ref_data.update(
                    self._hashRef(Reference(m.b.b1[:].b_s[x1].v2[x2, j]))
                    for x1, x2, j in SSC
                )
                ref_data.update(self._hashRef(Reference(m.b.b2[:, x].v0)) for x in S)
                ref_data.update(
                    self._hashRef(Reference(m.b.b2[:, x].v1[j])) for x, j in SC
                )
                assert len(comps) == len(ref_data)
                for comp in comps:
                    self.assertIn(self._hashRef(comp), ref_data)
            elif len(sets) == 2 and sets[0] is m.time and sets[1] is m.time:
                ref_data = {self._hashRef(Reference(m.v_tt))}
                ref_data.update(self._hashRef(Reference(m.v_tst[:, x, :])) for x in S)
                ref_data.update(
                    self._hashRef(Reference(m.b.b2[:, x].v2[:, j])) for x, j in SC
                )
                assert len(comps) == len(ref_data)
                for comp in comps:
                    self.assertIn(self._hashRef(comp), ref_data)
            else:
                raise RuntimeError()

    def _model2_nd_sets(self):
        # A more simple model, but now with some higher-dimension sets
        m = ConcreteModel()

        normalize_index.flatten = False

        m.d1 = Set(initialize=[1, 2])
        m.d2 = Set(initialize=[('a', 1), ('b', 2)])
        m.dn = Set(initialize=[('c', 3), ('d', 4, 5)], dimen=None)

        m.v_2n = Var(m.d2, m.dn)
        m.v_12 = Var(m.d1, m.d2)
        m.v_212 = Var(m.d2, m.d1, m.d2)
        m.v_12n = Var(m.d1, m.d2, m.dn)
        m.v_1n2n = Var(m.d1, m.dn, m.d2, m.dn)

        @m.Block(m.d1, m.d2, m.dn)
        def b(b, i1, i2, i3):
            b.v0 = Var()
            b.v1 = Var(m.d1)
            b.v2 = Var(m.d2)
            b.vn = Var(m.dn)

        normalize_index.flatten = True

        return m

    def test_flatten_m2_2d(self):
        """
        This test has some issues due to incompatibility between
        slicing and `normalize_index.flatten==False`.
        """
        # TODO: If the user wants to do what this test attempts
        # and "flatten along" sets of dimension > 1 when
        # `normalize_index.flatten == False`, they will have
        # some problems. See issue #1800.
        m = self._model2_nd_sets()

        sets = ComponentSet((m.d2,))
        # need to set `flatten` to False here to properly access data,
        # since model was created with `flatten == False`.
        normalize_index.flatten = False
        sets_list, comps_list = flatten_components_along_sets(m, sets, Var)

        ref1 = Reference(m.v_2n[:, ('c', 3)])

        ref_set = ref1.index_set()._ref  # _index is a _ReferenceSet

        # next(ref_set._get_iter(ref_set._slice, ('a',1)))
        # ^ This raises a somewhat cryptic error message.
        # _ReferenceSet.__contains__ seems to be the culprit,
        # which is called in validate_index for every __getitem__.
        self.assertNotIn(('a', 1), ref_set)
        # ^ This does not seem to behave as expected...
        # Reason is incompatibility with flatten==False.

        self.assertEqual(len(sets_list), len(comps_list))
        self.assertEqual(len(sets_list), 2)

        # for sets, comps in zip(sets_list, comps_list):
        #    if len(sets) == 1 and sets[0] is m.d2:
        #        ref_data = {
        #                self._hashRef(Reference(m.v_2n[:,('a',1)])),
        #                self._hashRef(Reference(m.v_2n[:,('b',2)])),
        #                }
        #        # Cannot access the data of ^these references
        #        for comp in comps:
        #            self.assertIn(self._hashRef(comp), ref_data)

        normalize_index.flatten = True

    def test_flatten_m2_1d(self):
        m = self._model2_nd_sets()

        sets = ComponentSet((m.d1,))
        # need to set `flatten` to False here to properly access data,
        # since model was created with `flatten == False`.
        normalize_index.flatten = False
        sets_list, comps_list = flatten_components_along_sets(m, sets, Var)

        assert len(sets_list) == len(comps_list)
        assert len(sets_list) == 3

        D22 = m.d2 * m.d2
        D2N = m.d2 * m.dn
        DN2N = m.dn.cross(m.d2, m.dn)
        D2NN = m.d2.cross(m.dn, m.dn)
        D2N2 = m.d2.cross(m.dn, m.d2)

        for sets, comps in zip(sets_list, comps_list):
            if len(sets) == 1 and sets[0] is m.d1:
                ref_data = set()
                # Don't expand indices:
                ref_data.update(self._hashRef(Reference(m.v_12[:, i2])) for i2 in m.d2)
                ref_data.update(
                    self._hashRef(Reference(m.v_212[i2a, :, i2b])) for i2a, i2b in D22
                )
                ref_data.update(
                    self._hashRef(Reference(m.v_12n[:, i2, i_n])) for i2, i_n in D2N
                )
                ref_data.update(
                    self._hashRef(Reference(m.v_1n2n[:, i_na, i2, i_nb]))
                    for i_na, i2, i_nb in DN2N
                )
                ref_data.update(
                    self._hashRef(Reference(m.b[:, i2, i_n].v0)) for i2, i_n in D2N
                )
                ref_data.update(
                    self._hashRef(Reference(m.b[:, i2a, i_n].v2[i2b]))
                    for i2a, i_n, i2b in D2N2
                )
                ref_data.update(
                    self._hashRef(Reference(m.b[:, i2, i_na].vn[i_nb]))
                    for i2, i_na, i_nb in D2NN
                )
                # Expect length to be 38
                assert len(ref_data) == len(comps)
                for comp in comps:
                    self.assertIn(self._hashRef(comp), ref_data)
            elif len(sets) == 1 and sets[0] is UnindexedComponent_set:
                ref_data = set()
                ref_data.update(self._hashRef(v) for v in m.v_2n.values())
                assert len(ref_data) == len(comps)
                for comp in comps:
                    self.assertIn(self._hashRef(comp), ref_data)
            elif len(sets) == 2 and sets[0] is m.d1 and sets[1] is m.d1:
                ref_data = set()
                ref_data.update(
                    self._hashRef(Reference(m.b[:, i2, i_n].v1[:])) for i2, i_n in D2N
                )
                assert len(ref_data) == len(comps)
                for comp in comps:
                    self.assertIn(self._hashRef(comp), ref_data)
            else:
                raise RuntimeError()

        normalize_index.flatten = True

    def _model3_nd_sets_normalizeflatten(self):
        # The same as model 2, but now with `normalize_index.flatten == True`
        m = ConcreteModel()

        m.d1 = Set(initialize=[1, 2])
        m.d2 = Set(initialize=[('a', 1), ('b', 2)])
        m.dn = Set(initialize=[('c', 3), ('d', 4, 5)], dimen=None)

        m.v_2n = Var(m.d2, m.dn)
        m.v_12 = Var(m.d1, m.d2)
        m.v_212 = Var(m.d2, m.d1, m.d2)
        m.v_12n = Var(m.d1, m.d2, m.dn)
        m.v_1n2n = Var(m.d1, m.dn, m.d2, m.dn)

        m.b = Block(m.d1, m.d2, m.dn)
        for i1 in m.d1:
            for i2 in m.d2:
                for i_n in m.dn:
                    m.b[i1, i2, i_n].v0 = Var()
                    m.b[i1, i2, i_n].v1 = Var(m.d1)
                    m.b[i1, i2, i_n].v2 = Var(m.d2)
                    m.b[i1, i2, i_n].vn = Var(m.dn)

        return m

    def test_flatten_m3_1d(self):
        m = self._model3_nd_sets_normalizeflatten()

        sets = ComponentSet((m.d1,))
        sets_list, comps_list = flatten_components_along_sets(m, sets, Var)

        assert len(sets_list) == len(comps_list)
        assert len(sets_list) == 3

        for sets, comps in zip(sets_list, comps_list):
            if len(sets) == 1 and sets[0] is m.d1:
                ref_data = set()
                # Must iterate and slice in a manner consistent with
                # `normalize_index.flatten == True`
                ref_data.update(
                    self._hashRef(Reference(m.v_12[:, i2])) for i2 in m.d2
                )  # 2
                ref_data.update(
                    self._hashRef(Reference(m.v_212[i2a, :, i2b]))
                    for i2a in m.d2
                    for i2b in m.d2
                )  # 4
                ref_data.update(
                    self._hashRef(Reference(m.v_12n[:, i2, i_n]))
                    for i2 in m.d2
                    for i_n in m.dn
                )  # 4
                ref_data.update(
                    self._hashRef(Reference(m.v_1n2n[:, i_na, i2, i_nb]))
                    for i_na in m.dn
                    for i2 in m.d2
                    for i_nb in m.dn
                )  # 8
                ref_data.update(
                    self._hashRef(Reference(m.b[:, i2, i_n].v0))
                    for i2 in m.d2
                    for i_n in m.dn
                )  # 4
                ref_data.update(
                    self._hashRef(Reference(m.b[:, i2a, i_n].v2[i2b]))
                    for i2a in m.d2
                    for i_n in m.dn
                    for i2b in m.d2
                )  # 8
                ref_data.update(
                    self._hashRef(Reference(m.b[:, i2, i_na].vn[i_nb]))
                    for i2 in m.d2
                    for i_na in m.dn
                    for i_nb in m.dn
                )  # 8
                assert len(ref_data) == len(comps)
                assert len(ref_data) == 38
                for comp in comps:
                    self.assertIn(self._hashRef(comp), ref_data)
            elif len(sets) == 1 and sets[0] is UnindexedComponent_set:
                ref_data = set()
                ref_data.update(self._hashRef(v) for v in m.v_2n.values())
                assert len(ref_data) == len(comps)
                for comp in comps:
                    self.assertIn(self._hashRef(comp), ref_data)
            elif len(sets) == 2 and sets[0] is m.d1 and sets[1] is m.d1:
                ref_data = set()
                ref_data.update(
                    self._hashRef(Reference(m.b[:, i2, i_n].v1[:]))
                    for i2 in m.d2
                    for i_n in m.dn
                )
                assert len(ref_data) == len(comps)
                for comp in comps:
                    self.assertIn(self._hashRef(comp), ref_data)

    def test_flatten_m3_2d(self):
        m = self._model3_nd_sets_normalizeflatten()

        sets = ComponentSet((m.d2,))
        sets_list, comps_list = flatten_components_along_sets(m, sets, Var)

        assert len(sets_list) == len(comps_list)
        assert len(sets_list) == 2

        for sets, comps in zip(sets_list, comps_list):
            if len(sets) == 1 and sets[0] is m.d2:
                ref_data = set()
                ref_data.update(
                    self._hashRef(Reference(m.v_2n[:, :, i_n])) for i_n in m.dn
                )  # 2
                ref_data.update(
                    self._hashRef(Reference(m.v_12[i1, :, :])) for i1 in m.d1
                )  # 2
                ref_data.update(
                    self._hashRef(Reference(m.v_12n[i1, :, :, i_n]))
                    for i1 in m.d1
                    for i_n in m.dn
                )  # 4
                ref_data.update(
                    self._hashRef(Reference(m.v_1n2n[i1, i_na, :, :, i_nb]))
                    for i1 in m.d1
                    for i_na in m.dn
                    for i_nb in m.dn
                )  # 8
                ref_data.update(
                    self._hashRef(Reference(m.b[i1, :, :, i_n].v0))
                    for i1 in m.d1
                    for i_n in m.dn
                )  # 4
                ref_data.update(
                    self._hashRef(Reference(m.b[i1a, :, :, i_n].v1[i1b]))
                    for i1a in m.d1
                    for i_n in m.dn
                    for i1b in m.d1
                )  # 8
                ref_data.update(
                    self._hashRef(Reference(m.b[i1, :, :, i_na].vn[i_nb]))
                    for i1 in m.d1
                    for i_na in m.dn
                    for i_nb in m.dn
                )  # 8
                assert len(ref_data) == len(comps)
                assert len(ref_data) == 36
                for comp in comps:
                    self.assertIn(self._hashRef(comp), ref_data)
            elif len(sets) == 2 and sets[0] is m.d2 and sets[1] is m.d2:
                ref_data = set()
                ref_data.update(
                    self._hashRef(Reference(m.v_212[:, :, i1, :, :])) for i1 in m.d1
                )
                ref_data.update(
                    self._hashRef(Reference(m.b[i1, :, :, i_n].v2[:, :]))
                    for i1 in m.d1
                    for i_n in m.dn
                )
                assert len(ref_data) == len(comps)
                assert len(ref_data) == 6
                for comp in comps:
                    self.assertIn(self._hashRef(comp), ref_data)
            else:
                raise RuntimeError()

    def test_flatten_m3_nd(self):
        m = self._model3_nd_sets_normalizeflatten()

        # Can't create a slice with multiple ellipses in the same index.
        m.del_component(m.v_1n2n)

        sets = ComponentSet((m.dn,))
        sets_list, comps_list = flatten_components_along_sets(m, sets, Var)

        assert len(sets_list) == len(comps_list)
        assert len(sets_list) == 3

        for sets, comps in zip(sets_list, comps_list):
            if len(sets) == 1 and sets[0] is UnindexedComponent_set:
                ref_data = set()
                ref_data.update(self._hashRef(v) for v in m.v_12.values())
                ref_data.update(self._hashRef(v) for v in m.v_212.values())
                assert len(comps) == len(ref_data)
                assert len(comps) == 12
                for comp in comps:
                    self.assertIn(self._hashRef(comp), ref_data)

            elif len(sets) == 1 and sets[0] is m.dn:
                ref_data = set()
                ref_data.update(
                    self._hashRef(Reference(m.v_2n[i2, ...])) for i2 in m.d2
                )  # 2
                ref_data.update(
                    self._hashRef(Reference(m.v_12n[i1, i2, ...]))
                    for i1 in m.d1
                    for i2 in m.d2
                )  # 4
                ref_data.update(
                    self._hashRef(Reference(m.b[i1, i2, ...].v0))
                    for i1 in m.d1
                    for i2 in m.d2
                )  # 4
                ref_data.update(
                    self._hashRef(Reference(m.b[i1a, i2, ...].v1[i1b]))
                    for i1a in m.d1
                    for i2 in m.d2
                    for i1b in m.d1
                )  # 8
                ref_data.update(
                    self._hashRef(Reference(m.b[i1, i2a, ...].v2[i2b]))
                    for i1 in m.d1
                    for i2a in m.d2
                    for i2b in m.d2
                )  # 8
                assert len(comps) == len(ref_data)
                assert len(comps) == 26
                for comp in comps:
                    self.assertIn(self._hashRef(comp), ref_data)

            elif len(sets) == 2 and sets[0] is m.dn and sets[1] is m.dn:
                ref_data = set()
                ref_data.update(
                    self._hashRef(Reference(m.b[i1, i2, ...].vn[...]))
                    for i1 in m.d1
                    for i2 in m.d2
                )  # 4
                assert len(comps) == len(ref_data)
                for comp in comps:
                    self.assertIn(self._hashRef(comp), ref_data)
            else:
                raise RuntimeError()

    def test_flatten_m3_1_2(self):
        m = self._model3_nd_sets_normalizeflatten()

        sets = ComponentSet((m.d1, m.d2))
        sets_list, comps_list = flatten_components_along_sets(m, sets, Var)

        assert len(sets_list) == len(comps_list)
        assert len(sets_list) == 5

        for sets, comps in zip(sets_list, comps_list):
            if len(sets) == 1 and sets[0] is m.d2:
                ref_data = set()
                ref_data.update(
                    self._hashRef(Reference(m.v_2n[:, :, i_n])) for i_n in m.dn
                )
                self.assertEqual(len(ref_data), len(comps))
                for comp in comps:
                    self.assertIn(self._hashRef(comp), ref_data)

            elif len(sets) == 2 and sets[0] is m.d1 and sets[1] is m.d2:
                ref_data = {self._hashRef(Reference(m.v_12[...]))}  # 1
                ref_data.update(
                    self._hashRef(Reference(m.v_12n[:, :, :, i_n])) for i_n in m.dn
                )  # 2
                ref_data.update(
                    self._hashRef(Reference(m.v_1n2n[:, i_na, :, :, i_nb]))
                    for i_na in m.dn
                    for i_nb in m.dn
                )  # 4
                ref_data.update(
                    self._hashRef(Reference(m.b[:, :, :, i_n].v0)) for i_n in m.dn
                )  # 2
                ref_data.update(
                    self._hashRef(Reference(m.b[:, :, :, i_na].vn[i_nb]))
                    for i_na in m.dn
                    for i_nb in m.dn
                )  # 4
                self.assertEqual(len(ref_data), len(comps))
                self.assertEqual(len(comps), 13)
                for comp in comps:
                    self.assertIn(self._hashRef(comp), ref_data)

            elif (
                len(sets) == 3
                and sets[0] is m.d1
                and sets[1] is m.d2
                and sets[2] is m.d1
            ):
                ref_data = set()
                ref_data.update(
                    self._hashRef(Reference(m.b[:, :, :, i_n].v1[:])) for i_n in m.dn
                )  # 2
                self.assertEqual(len(ref_data), len(comps))
                for comp in comps:
                    self.assertIn(self._hashRef(comp), ref_data)

            elif (
                len(sets) == 3
                and sets[0] is m.d1
                and sets[1] is m.d2
                and sets[2] is m.d2
            ):
                ref_data = set()
                ref_data.update(
                    self._hashRef(Reference(m.b[:, :, :, i_n].v2[:, :])) for i_n in m.dn
                )  # 2
                self.assertEqual(len(ref_data), len(comps))
                for comp in comps:
                    self.assertIn(self._hashRef(comp), ref_data)

            elif (
                len(sets) == 3
                and sets[0] is m.d2
                and sets[1] is m.d1
                and sets[2] is m.d2
            ):
                ref_data = {self._hashRef(Reference(m.v_212[...]))}  # 1
                self.assertEqual(len(ref_data), len(comps))
                for comp in comps:
                    self.assertIn(self._hashRef(comp), ref_data)

            else:
                raise RuntimeError()

    def test_specified_index_1(self):
        """
        Components indexed by flattened sets and others
        """
        m = ConcreteModel()

        m.time = Set(initialize=[1, 2, 3])
        m.space = Set(initialize=[2, 4, 6])
        m.phase = Set(initialize=['p1', 'p2'])
        m.comp = Set(initialize=['a', 'b'])

        phase_comp = m.comp * m.phase
        n_phase_comp = len(m.phase) * len(m.comp)

        m.v = Var(m.time, m.comp, m.space, m.phase)

        @m.Block(m.time, m.comp, m.space, m.phase)
        def b(b, t, j, x, p):
            b.v1 = Var()

            if x != 2:
                b.v2 = Var()

        sets = (m.time, m.space)

        # No specified indices
        sets_list, comps_list = flatten_components_along_sets(m, sets, Var)

        # Everything is indexed by time and space
        self.assertEqual(len(comps_list), len(sets_list))
        self.assertEqual(len(sets_list), 1)

        for sets, comps in zip(sets_list, comps_list):
            if len(sets) == 2 and sets[0] is m.time and sets[1] is m.space:
                # We missed b.v2 by descending into the "first" index
                # of the block
                self.assertEqual(len(comps), 2 * n_phase_comp)
                ref_data = set()
                ref_data.update(
                    self._hashRef(Reference(m.v[:, j, :, p])) for j, p in phase_comp
                )
                ref_data.update(
                    self._hashRef(Reference(m.b[:, j, :, p].v1)) for j, p in phase_comp
                )
                self.assertEqual(len(ref_data), len(comps))
                for comp in comps:
                    self.assertIn(self._hashRef(comp), ref_data)
            else:
                raise RuntimeError()

        # Space index specified:
        indices = ComponentMap([(m.space, 4)])
        sets_list, comps_list = flatten_components_along_sets(
            m, sets, Var, indices=indices
        )

        self.assertEqual(len(comps_list), len(sets_list))
        self.assertEqual(len(sets_list), 1)

        for sets, comps in zip(sets_list, comps_list):
            if len(sets) == 2 and sets[0] is m.time and sets[1] is m.space:
                # We descended into a block data that includes v2
                self.assertEqual(len(comps), 3 * n_phase_comp)

                # Slices where we expect an attribute error somewhere,
                # due to v2 being "skipped"
                incomplete_slices = list(m.b[:, j, :, p].v2 for j, p in phase_comp)
                for ref in incomplete_slices:
                    ref.attribute_errors_generate_exceptions = False
                incomplete_refs = list(Reference(sl) for sl in incomplete_slices)

                ref_data = set()
                ref_data.update(
                    self._hashRef(Reference(m.v[:, j, :, p])) for j, p in phase_comp
                )
                ref_data.update(
                    self._hashRef(Reference(m.b[:, j, :, p].v1)) for j, p in phase_comp
                )
                ref_data.update(self._hashRef(ref) for ref in incomplete_refs)
                self.assertEqual(len(ref_data), len(comps))
                for comp in comps:
                    self.assertIn(self._hashRef(comp), ref_data)
            else:
                raise RuntimeError()

        # Time and space indices specified
        indices = (3, 6)
        sets_list, comps_list = flatten_components_along_sets(
            m, sets, Var, indices=indices
        )

        self.assertEqual(len(comps_list), len(sets_list))
        self.assertEqual(len(sets_list), 1)

        for sets, comps in zip(sets_list, comps_list):
            if len(sets) == 2 and sets[0] is m.time and sets[1] is m.space:
                # We descended into a block data that includes v2
                self.assertEqual(len(comps), 3 * n_phase_comp)

                # Slices where we expect an attribute error somewhere,
                # due to v2 being "skipped"
                incomplete_slices = list(m.b[:, j, :, p].v2 for j, p in phase_comp)
                for ref in incomplete_slices:
                    ref.attribute_errors_generate_exceptions = False
                incomplete_refs = list(Reference(sl) for sl in incomplete_slices)

                ref_data = set()
                ref_data.update(
                    self._hashRef(Reference(m.v[:, j, :, p])) for j, p in phase_comp
                )
                ref_data.update(
                    self._hashRef(Reference(m.b[:, j, :, p].v1)) for j, p in phase_comp
                )
                ref_data.update(self._hashRef(ref) for ref in incomplete_refs)
                self.assertEqual(len(ref_data), len(comps))
                for comp in comps:
                    self.assertIn(self._hashRef(comp), ref_data)
            else:
                raise RuntimeError()

    def test_specified_index_2(self):
        """
        Components indexed only by flattened sets
        """
        m = ConcreteModel()

        m.time = Set(initialize=[1, 2, 3])
        m.space = Set(initialize=[2, 4, 6])

        m.v = Var(m.time, m.space)

        @m.Block(m.time, m.space)
        def b(b, t, x):
            b.v1 = Var()

            if x != 2:
                b.v2 = Var()

        sets = (m.time, m.space)

        # No specified indices
        sets_list, comps_list = flatten_components_along_sets(m, sets, Var)

        # Everything is indexed by time and space
        self.assertEqual(len(comps_list), len(sets_list))
        self.assertEqual(len(sets_list), 1)

        for sets, comps in zip(sets_list, comps_list):
            if len(sets) == 2 and sets[0] is m.time and sets[1] is m.space:
                # We missed b.v2 by descending into the "first" index
                # of the block
                self.assertEqual(len(comps), 2)
                ref_data = {
                    self._hashRef(Reference(m.v[...])),
                    self._hashRef(Reference(m.b[...].v1)),
                }
                self.assertEqual(len(ref_data), len(comps))
                for comp in comps:
                    self.assertIn(self._hashRef(comp), ref_data)
            else:
                raise RuntimeError()

        # Space index specified:
        indices = ComponentMap([(m.space, 4)])
        sets_list, comps_list = flatten_components_along_sets(
            m, sets, Var, indices=indices
        )

        self.assertEqual(len(comps_list), len(sets_list))
        self.assertEqual(len(sets_list), 1)

        for sets, comps in zip(sets_list, comps_list):
            if len(sets) == 2 and sets[0] is m.time and sets[1] is m.space:
                # We descended into a block data that includes v2
                self.assertEqual(len(comps), 3)

                # Slices where we expect an attribute error somewhere,
                # due to v2 being "skipped"
                incomplete_slice = m.b[:, :].v2
                incomplete_slice.attribute_errors_generate_exceptions = False
                incomplete_ref = Reference(incomplete_slice)

                ref_data = {
                    self._hashRef(Reference(m.v[:, :])),
                    self._hashRef(Reference(m.b[:, :].v1)),
                    self._hashRef(incomplete_ref),
                }
                self.assertEqual(len(ref_data), len(comps))
                for comp in comps:
                    self.assertIn(self._hashRef(comp), ref_data)
            else:
                raise RuntimeError()

        # Time and space indices specified
        indices = (3, 6)
        sets_list, comps_list = flatten_components_along_sets(
            m, sets, Var, indices=indices
        )

        self.assertEqual(len(comps_list), len(sets_list))
        self.assertEqual(len(sets_list), 1)

        for sets, comps in zip(sets_list, comps_list):
            if len(sets) == 2 and sets[0] is m.time and sets[1] is m.space:
                # We descended into a block data that includes v2
                self.assertEqual(len(comps), 3)

                # Slices where we expect an attribute error somewhere,
                # due to v2 being "skipped"
                incomplete_slice = m.b[:, :].v2
                incomplete_slice.attribute_errors_generate_exceptions = False
                incomplete_ref = Reference(incomplete_slice)

                ref_data = {
                    self._hashRef(Reference(m.v[:, :])),
                    self._hashRef(Reference(m.b[:, :].v1)),
                    self._hashRef(incomplete_ref),
                }
                self.assertEqual(len(ref_data), len(comps))
                for comp in comps:
                    self.assertIn(self._hashRef(comp), ref_data)
            else:
                raise RuntimeError()

    def _model4_three_1d_sets(self):
        # A simple model with three sets to slice
        m = ConcreteModel()

        m.X = Set(initialize=[1, 2, 3])
        m.Y = Set(initialize=[1, 2, 3])
        m.Z = Set(initialize=[1, 2, 3])

        m.comp = Set(initialize=['a', 'b'])

        m.u = Var()

        m.v = Var(m.X, m.Y, m.Z, m.comp)
        m.base = Var(m.X, m.Y)

        @m.Block(m.X, m.Y, m.Z, m.comp)
        def b4(b, x, y, z, j):
            b.v = Var()

        @m.Block(m.X, m.Y)
        def b2(b, x, y):
            b.base = Var()
            b.v = Var(m.Z, m.comp)

        return m

    def test_model4_xyz(self):
        m = self._model4_three_1d_sets()

        sets = (m.X, m.Y, m.Z)
        sets_list, comps_list = flatten_components_along_sets(m, sets, Var)

        self.assertEqual(len(comps_list), len(sets_list))
        self.assertEqual(len(sets_list), 3)

        for sets, comps in zip(sets_list, comps_list):
            if len(sets) == 1 and sets[0] is UnindexedComponent_set:
                ref_data = {self._hashRef(Reference(m.u))}
                self.assertEqual(len(ref_data), len(comps))
                for comp in comps:
                    self.assertIn(self._hashRef(comp), ref_data)
            elif len(sets) == 2 and sets[0] is m.X and sets[1] is m.Y:
                ref_data = {
                    self._hashRef(Reference(m.base[:, :])),
                    self._hashRef(Reference(m.b2[:, :].base)),
                }
                self.assertEqual(len(ref_data), len(comps))
                for comp in comps:
                    self.assertIn(self._hashRef(comp), ref_data)
            elif (
                len(sets) == 3 and sets[0] is m.X and sets[1] is m.Y and sets[2] is m.Z
            ):
                ref_data = {
                    self._hashRef(Reference(m.v[:, :, :, 'a'])),
                    self._hashRef(Reference(m.v[:, :, :, 'b'])),
                    self._hashRef(Reference(m.b4[:, :, :, 'a'].v)),
                    self._hashRef(Reference(m.b4[:, :, :, 'b'].v)),
                    self._hashRef(Reference(m.b2[:, :].v[:, 'a'])),
                    self._hashRef(Reference(m.b2[:, :].v[:, 'b'])),
                }
                self.assertEqual(len(ref_data), len(comps))
                for comp in comps:
                    self.assertIn(self._hashRef(comp), ref_data)
            else:
                raise RuntimeError()

    def test_deactivated_block_active_true(self):
        m = self._model1_1d_sets()

        # Deactivating b1 should get rid of both variables directly on it
        # as well as those on the subblock b_s
        m.b.b1.deactivate()
        sets = (m.time,)

        #
        # Test identifying active components
        #
        sets_list, comps_list = flatten_components_along_sets(m, sets, Var, active=True)

        expected_unindexed = [ComponentUID(m.v0)]
        expected_unindexed = set(expected_unindexed)
        expected_time = [ComponentUID(m.v1[:])]
        expected_time.extend(ComponentUID(m.v2[:, x]) for x in m.space)
        expected_time.extend(
            ComponentUID(m.v3[:, x, j]) for x in m.space for j in m.comp
        )
        expected_time.extend(ComponentUID(m.b.b2[:, x].v0) for x in m.space)
        expected_time.extend(
            ComponentUID(m.b.b2[:, x].v1[j]) for x in m.space for j in m.comp
        )
        expected_time = set(expected_time)

        expected_2time = [ComponentUID(m.v_tt[:, :])]
        expected_2time.extend(ComponentUID(m.v_tst[:, x, :]) for x in m.space)
        expected_2time.extend(
            ComponentUID(m.b.b2[:, x].v2[:, j]) for x in m.space for j in m.comp
        )
        expected_2time = set(expected_2time)

        set_id_set = set(tuple(id(s) for s in sets) for sets in sets_list)
        pred_sets = [(UnindexedComponent_set,), (m.time,), (m.time, m.time)]
        pred_set_ids = set(tuple(id(s) for s in sets) for sets in pred_sets)
        self.assertEqual(set_id_set, pred_set_ids)
        for sets, comps in zip(sets_list, comps_list):
            if len(sets) == 1 and sets[0] is UnindexedComponent_set:
                comp_set = set(ComponentUID(comp) for comp in comps)
                self.assertEqual(comp_set, expected_unindexed)
            elif len(sets) == 1 and sets[0] is m.time:
                comp_set = set(ComponentUID(comp.referent) for comp in comps)
                self.assertEqual(comp_set, expected_time)
            elif len(sets) == 2:
                self.assertIs(sets[0], m.time)
                self.assertIs(sets[1], m.time)
                comp_set = set(ComponentUID(comp.referent) for comp in comps)
                self.assertEqual(comp_set, expected_2time)

    def test_deactivated_block_active_false(self):
        m = self._model1_1d_sets()
        m.deactivate()
        m.b.deactivate()
        m.b.b1.deactivate()
        m.b.b1[:].b_s.deactivate()
        # Remove components to make this easier to test
        m.del_component(m.v0)
        m.del_component(m.v1)
        m.del_component(m.v3)
        m.del_component(m.v_tt)
        m.del_component(m.v_tst)
        sets = (m.time,)
        sets_list, comps_list = flatten_components_along_sets(
            m, sets, Var, active=False
        )

        expected_time = [ComponentUID(m.b.b1[:].v0)]
        expected_time.extend(ComponentUID(m.v2[:, x]) for x in m.space)
        expected_time.extend(ComponentUID(m.b.b1[:].v1[x]) for x in m.space)
        expected_time.extend(
            ComponentUID(m.b.b1[:].v2[x, j]) for x in m.space for j in m.comp
        )
        expected_time.extend(ComponentUID(m.b.b1[:].b_s[x].v0) for x in m.space)
        expected_time.extend(
            ComponentUID(m.b.b1[:].b_s[x1].v1[x2]) for x1 in m.space for x2 in m.space
        )
        expected_time.extend(
            ComponentUID(m.b.b1[:].b_s[x1].v2[x2, j])
            for x1 in m.space
            for x2 in m.space
            for j in m.comp
        )
        expected_time = set(expected_time)

        self.assertEqual(len(sets_list), 1)
        self.assertEqual(len(sets_list[0]), 1)
        self.assertIs(sets_list[0][0], m.time)

        self.assertEqual(len(comps_list), 1)
        comp_set = set(ComponentUID(comp.referent) for comp in comps_list[0])
        self.assertEqual(comp_set, expected_time)

    def test_partially_deactivated_slice_active_true(self):
        m = ConcreteModel()
        m.time = Set(initialize=[0, 1, 2, 3])
        m.b = Block(m.time)
        for t in m.time:
            m.b[t].v = Var()
        m.b[0].deactivate()
        m.b[1].deactivate()
        # m.b[:] is now a "partially deactivated slice"
        sets = (m.time,)
        sets_list, comps_list = flatten_components_along_sets(m, sets, Var, active=True)
        self.assertEqual(len(sets_list), 1)
        self.assertEqual(len(sets_list[0]), 1)
        self.assertIs(sets_list[0][0], m.time)
        self.assertIs(sets_list[0][0], m.time)
        self.assertEqual(len(comps_list), 1)
        self.assertEqual(len(comps_list[0]), 1)
        self.assertEqual(
            ComponentUID(comps_list[0][0].referent), ComponentUID(m.b[:].v)
        )

    def test_partially_activated_slice_active_false(self):
        m = ConcreteModel()
        m.time = Set(initialize=[0, 1, 2, 3])
        m.b = Block(m.time)
        m.deactivate()
        m.b.deactivate()
        for t in m.time:
            m.b[t].v = Var()
        m.b[0].deactivate()
        m.b[1].deactivate()
        # Note that m.b[2] and m.b[3] are active.
        # m.b[:] is now a "partially activated slice"
        sets = (m.time,)
        sets_list, comps_list = flatten_components_along_sets(
            m, sets, Var, active=False
        )
        self.assertEqual(len(sets_list), 1)
        self.assertEqual(len(sets_list[0]), 1)
        self.assertIs(sets_list[0][0], m.time)
        self.assertIs(sets_list[0][0], m.time)
        self.assertEqual(len(comps_list), 1)
        self.assertEqual(len(comps_list[0]), 1)
        self.assertEqual(
            ComponentUID(comps_list[0][0].referent), ComponentUID(m.b[:].v)
        )

    def test_partially_deactivated_slice_specified_index(self):
        m = ConcreteModel()
        m.time = Set(initialize=[0, 1, 2, 3])
        m.b = Block(m.time)
        for t in m.time:
            m.b[t].v = Var()
        m.b[0].deactivate()
        m.b[1].deactivate()
        # m.b[:] is now a "partially deactivated slice"
        sets = (m.time,)

        # When we specify the index of a deactivated block, we
        # respect the active argument when descending.
        indices = (1,)
        sets_list, comps_list = flatten_components_along_sets(
            m, sets, Var, active=True, indices=indices
        )
        self.assertEqual(len(sets_list), 0)
        self.assertEqual(len(comps_list), 0)

        indices = (2,)
        sets_list, comps_list = flatten_components_along_sets(
            m, sets, Var, active=True, indices=indices
        )
        self.assertEqual(len(sets_list), 1)
        self.assertEqual(len(sets_list[0]), 1)
        self.assertIs(sets_list[0][0], m.time)
        self.assertIs(sets_list[0][0], m.time)
        self.assertEqual(len(comps_list), 1)
        self.assertEqual(len(comps_list[0]), 1)
        self.assertEqual(
            ComponentUID(comps_list[0][0].referent), ComponentUID(m.b[:].v)
        )

    def test_fully_deactivated_slice(self):
        m = ConcreteModel()
        m.time = Set(initialize=[0, 1, 2, 3])
        m.b = Block(m.time)
        for t in m.time:
            m.b[t].v = Var()
        m.b[:].deactivate()
        sets = (m.time,)

        # We send active=True, but cannot find an active BlockData
        # to descend into.
        sets_list, comps_list = flatten_components_along_sets(m, sets, Var, active=True)
        self.assertEqual(len(sets_list), 0)
        self.assertEqual(len(comps_list), 0)

    def test_deactivated_model_active_false(self):
        m = self._model1_1d_sets()
        m.deactivate()
        sets = (m.time,)
        sets_list, comps_list = flatten_components_along_sets(m, sets, Var, active=True)
        self.assertEqual(len(sets_list), 0)
        self.assertEqual(len(comps_list), 0)

    def test_constraint_with_active_arg(self):
        m = ConcreteModel()
        m.time = Set(initialize=[0, 1, 2, 3])
        m.b = Block(m.time)
        for t in m.time:
            m.b[t].v = Var()
            m.b[t].c1 = Constraint(expr=m.b[t].v == 1)

        def c2_rule(m, t):
            return m.b[t].v == 2

        m.c2 = Constraint(m.time, rule=c2_rule)
        m.c2.deactivate()

        sets = (m.time,)
        sets_list, comps_list = flatten_components_along_sets(
            m, sets, Constraint, active=True
        )
        self.assertEqual(len(sets_list), 1)
        self.assertEqual(len(sets_list[0]), 1)
        self.assertIs(sets_list[0][0], m.time)
        self.assertIs(sets_list[0][0], m.time)
        self.assertEqual(len(comps_list), 1)
        self.assertEqual(len(comps_list[0]), 1)
        self.assertEqual(
            ComponentUID(comps_list[0][0].referent), ComponentUID(m.b[:].c1)
        )

        m.deactivate()
        sets = (m.time,)
        sets_list, comps_list = flatten_components_along_sets(
            m, sets, Constraint, active=False
        )
        self.assertEqual(len(sets_list), 1)
        self.assertEqual(len(sets_list[0]), 1)
        self.assertIs(sets_list[0][0], m.time)
        self.assertIs(sets_list[0][0], m.time)
        self.assertEqual(len(comps_list), 1)
        self.assertEqual(len(comps_list[0]), 1)
        self.assertEqual(ComponentUID(comps_list[0][0].referent), ComponentUID(m.c2[:]))

    def test_constraint_partially_deactivated_slice(self):
        m = ConcreteModel()
        m.time = Set(initialize=[0, 1, 2, 3])
        m.b = Block(m.time)
        for t in m.time:
            m.b[t].v = Var()

        def c2_rule(m, t):
            return m.b[t].v == 2

        m.c2 = Constraint(m.time, rule=c2_rule)
        m.c2[0].deactivate()
        m.c2[1].deactivate()

        sets = (m.time,)
        sets_list, comps_list = flatten_components_along_sets(
            m, sets, Constraint, active=True
        )
        self.assertEqual(len(sets_list), 1)
        self.assertEqual(len(sets_list[0]), 1)
        self.assertIs(sets_list[0][0], m.time)
        self.assertIs(sets_list[0][0], m.time)
        self.assertEqual(len(comps_list), 1)
        self.assertEqual(len(comps_list[0]), 1)
        self.assertEqual(ComponentUID(comps_list[0][0].referent), ComponentUID(m.c2[:]))

    def test_constraint_fully_deactivated_slice(self):
        m = ConcreteModel()
        m.time = Set(initialize=[0, 1, 2, 3])
        m.b = Block(m.time)
        for t in m.time:
            m.b[t].v = Var()

        def c2_rule(m, t):
            return m.b[t].v == 2

        m.c2 = Constraint(m.time, rule=c2_rule)
        m.c2[:].deactivate()

        sets = (m.time,)
        sets_list, comps_list = flatten_components_along_sets(
            m, sets, Constraint, active=True
        )
        # Because all data objects in c2[:] are deactivated, we don't
        # yield the slice.
        self.assertEqual(len(sets_list), 0)
        self.assertEqual(len(comps_list), 0)

    def test_scalar_con_active_true(self):
        m = ConcreteModel()
        m.time = Set(initialize=[0, 1, 2])
        m.v = Var()
        m.c = Constraint(expr=m.v == 1)
        sets = (m.time,)
        sets_list, comps_list = flatten_components_along_sets(
            m, sets, Constraint, active=True
        )

        self.assertEqual(len(sets_list), 1)
        self.assertEqual(len(sets_list[0]), 1)
        self.assertIs(sets_list[0][0], UnindexedComponent_set)
        self.assertEqual(len(comps_list), 1)
        self.assertEqual(len(comps_list[0]), 1)
        self.assertIs(comps_list[0][0], m.c)

    def test_deactivated_scalar_con_active_true(self):
        m = ConcreteModel()
        m.time = Set(initialize=[0, 1, 2])
        m.comp = Set(initialize=["A", "B"])
        m.v = Var()

        def c_rule(m, j):
            return m.v == 1

        m.c = Constraint(m.comp, rule=c_rule)
        m.c[:].deactivate()
        # Because only the data objects are deactivated, we will generate
        # this component in the component_objects loop in the flattener.
        # But because its data objects do not match the active argument,
        # we hit the clause that checks for slice activity. This checks
        # the part of the clause that makes sure we have sliced a set.
        sets = (m.time,)
        sets_list, comps_list = flatten_components_along_sets(
            m, sets, Constraint, active=True
        )

        self.assertEqual(len(sets_list), 0)
        self.assertEqual(len(comps_list), 0)


class TestCUID(unittest.TestCase):
    """
    When returning indexed components, the flattener returns references.
    Unless these are subsequently attached to a model (and maybe even if they
    are), these references will not have useful names. Creating a CUID
    from the referent attribute of these references is the preferred way
    to generate these names, because these names will be unique if these
    references are generated multiple times.

    However, when referring to a slice, a CUID is not truly unique, as
    "m.b[*].v" is often equivalent to "m.b[**].v".
    Our convention is to always use constant-dimension slices ("*")
    unless we are slicing a component with a None-dimensioned set.

    These tests assert that we use the correct convention.

    """

    # 3 cases to cover:
    # Components indexed by no sets we're interested in
    # Components indexed by some sets we're interested in
    # Components indexed by all sets we're interested in

    def test_cuids_no_sets_no_subblocks(self):
        m = ConcreteModel()
        m.s1 = Set(initialize=[1, 2, 3])
        m.s2 = Set(initialize=["a", "b"])
        m.s3 = Set(initialize=[4, 5, 6])
        m.s4 = Set(initialize=["c", "d"])
        m.v1 = Var(m.s3, m.s4)

        pred_cuid_set = {
            "v1[4,c]",
            "v1[4,d]",
            "v1[5,c]",
            "v1[5,d]",
            "v1[6,c]",
            "v1[6,d]",
        }

        sets = (m.s1, m.s2)
        ctype = Var
        sets_list, comps_list = flatten_components_along_sets(m, sets, ctype)
        for sets, comps in zip(sets_list, comps_list):
            if len(sets) == 1 and sets[0] is UnindexedComponent_set:
                self.assertEqual(len(comps), len(m.s1) * len(m.s2))
                cuid_set = set(str(ComponentUID(comp)) for comp in comps)
                self.assertEqual(cuid_set, pred_cuid_set)
            else:
                raise RuntimeError()

    def test_cuids_some_sets_no_subblocks(self):
        m = ConcreteModel()
        m.s1 = Set(initialize=[1, 2, 3])
        m.s2 = Set(initialize=["a", "b"])
        m.s3 = Set(initialize=[4, 5, 6])
        m.s4 = Set(initialize=["c", "d"])
        m.v1 = Var(m.s1, m.s4)

        pred_cuid_set = {"v1[1,*]", "v1[2,*]", "v1[3,*]"}

        sets = (m.s3, m.s4)
        ctype = Var
        sets_list, comps_list = flatten_components_along_sets(m, sets, ctype)
        for sets, comps in zip(sets_list, comps_list):
            if len(sets) == 1 and sets[0] is m.s4:
                self.assertEqual(len(comps), len(m.s1))
                cuid_set = set(str(ComponentUID(comp.referent)) for comp in comps)
                self.assertEqual(cuid_set, pred_cuid_set)
            else:
                raise RuntimeError()

    def test_cuids_all_sets_no_subblocks(self):
        m = ConcreteModel()
        m.s1 = Set(initialize=[1, 2, 3])
        m.s2 = Set(initialize=["a", "b"])
        m.s3 = Set(initialize=[4, 5, 6])
        m.s4 = Set(initialize=["c", "d"])
        m.v1 = Var(m.s3, m.s4)

        pred_cuid_set = {"v1[*,*]"}

        sets = (m.s3, m.s4)
        ctype = Var
        sets_list, comps_list = flatten_components_along_sets(m, sets, ctype)
        for sets, comps in zip(sets_list, comps_list):
            if len(sets) == 2 and sets[0] is m.s3 and sets[1] is m.s4:
                self.assertEqual(len(comps), 1)
                cuid_set = set(str(ComponentUID(comp.referent)) for comp in comps)
                self.assertEqual(cuid_set, pred_cuid_set)
            else:
                raise RuntimeError()

    def test_cuid_one_set_no_subblocks(self):
        m = ConcreteModel()
        m.s1 = Set(initialize=[1, 2, 3])
        m.v = Var(m.s1)

        pred_cuid_set = {"v[*]"}

        sets = (m.s1,)
        ctype = Var
        sets_list, comps_list = flatten_components_along_sets(m, sets, ctype)
        self.assertEqual(len(sets_list), 1)
        self.assertEqual(len(comps_list), 1)
        for sets, comps in zip(sets_list, comps_list):
            if len(sets) == 1 and sets[0] is m.s1:
                self.assertEqual(len(comps), 1)
                cuid_set = set(str(ComponentUID(comp.referent)) for comp in comps)
                self.assertEqual(cuid_set, pred_cuid_set)
            else:
                raise RuntimeError()

    def test_cuids_no_sets_with_subblocks(self):
        m = ConcreteModel()
        m.s1 = Set(initialize=[1, 2, 3])
        m.s2 = Set(initialize=["a", "b"])
        m.s3 = Set(initialize=[4, 5, 6])
        m.s4 = Set(initialize=["c", "d"])

        def block_rule(b, i, j):
            b.v = Var()

        m.b = Block(m.s1, m.s2, rule=block_rule)

        pred_cuid_set = {
            "b[1,a].v",
            "b[1,b].v",
            "b[2,a].v",
            "b[2,b].v",
            "b[3,a].v",
            "b[3,b].v",
        }

        sets = (m.s3, m.s4)
        ctype = Var
        sets_list, comps_list = flatten_components_along_sets(m, sets, ctype)
        self.assertEqual(len(sets_list), 1)
        self.assertEqual(len(comps_list), 1)
        for sets, comps in zip(sets_list, comps_list):
            if len(sets) == 1 and sets[0] is UnindexedComponent_set:
                self.assertEqual(len(comps), len(m.s1) * len(m.s2))
                cuid_set = set(str(ComponentUID(comp)) for comp in comps)
                self.assertEqual(cuid_set, pred_cuid_set)
            else:
                raise RuntimeError()

    def test_cuids_some_sets_with_subblocks(self):
        m = ConcreteModel()
        m.s1 = Set(initialize=[1, 2, 3])
        m.s2 = Set(initialize=["a", "b"])
        m.s3 = Set(initialize=[4, 5, 6])
        m.s4 = Set(initialize=["c", "d"])

        def block_rule(b, i, j):
            b.v = Var()

        m.b = Block(m.s1, m.s2, rule=block_rule)

        pred_cuid_set = {"b[*,a].v", "b[*,b].v"}

        sets = (m.s1, m.s4)
        ctype = Var
        sets_list, comps_list = flatten_components_along_sets(m, sets, ctype)
        self.assertEqual(len(sets_list), 1)
        self.assertEqual(len(comps_list), 1)
        for sets, comps in zip(sets_list, comps_list):
            if len(sets) == 1 and sets[0] is m.s1:
                self.assertEqual(len(comps), len(m.s2))
                cuid_set = set(str(ComponentUID(comp.referent)) for comp in comps)
                self.assertEqual(cuid_set, pred_cuid_set)
            else:
                raise RuntimeError()

    def test_cuids_all_sets_with_subblocks(self):
        m = ConcreteModel()
        m.s1 = Set(initialize=[1, 2, 3])
        m.s2 = Set(initialize=["a", "b"])
        m.s3 = Set(initialize=[4, 5, 6])
        m.s4 = Set(initialize=["c", "d"])

        def block_rule(b, i, j):
            b.v = Var()

        m.b = Block(m.s1, m.s2, rule=block_rule)

        pred_cuid_set = {"b[*,*].v"}

        sets = (m.s1, m.s2)
        ctype = Var
        sets_list, comps_list = flatten_components_along_sets(m, sets, ctype)
        self.assertEqual(len(sets_list), 1)
        self.assertEqual(len(comps_list), 1)
        for sets, comps in zip(sets_list, comps_list):
            if len(sets) == 2 and sets[0] is m.s1 and sets[1] is m.s2:
                self.assertEqual(len(comps), 1)
                cuid_set = set(str(ComponentUID(comp.referent)) for comp in comps)
                self.assertEqual(cuid_set, pred_cuid_set)
            else:
                raise RuntimeError()

    def test_cuids_multiple_slices(self):
        m = ConcreteModel()
        m.s1 = Set(initialize=[1, 2, 3])

        def block_rule(b, i):
            b.v = Var(m.s1)

        m.b = Block(m.s1, rule=block_rule)

        pred_cuid_set = {"b[*].v[*]"}
        sets = (m.s1,)
        ctype = Var
        sets_list, comps_list = flatten_components_along_sets(m, sets, ctype)
        self.assertEqual(len(sets_list), 1)
        self.assertEqual(len(comps_list), 1)
        for sets, comps in zip(sets_list, comps_list):
            if len(sets) == 2 and sets[0] is m.s1 and sets[1] is m.s1:
                self.assertEqual(len(comps), 1)
                cuid_set = set(str(ComponentUID(comp.referent)) for comp in comps)
                self.assertEqual(cuid_set, pred_cuid_set)
            else:
                raise RuntimeError()


class TestSliceComponent(_TestFlattenBase, unittest.TestCase):
    def make_model(self):
        m = ConcreteModel()
        m.s1 = Set(initialize=[1, 2, 3])
        m.s2 = Set(initialize=["a", "b"])
        m.s3 = Set(initialize=[4, 5, 6])
        m.s4 = Set(initialize=["c", "d"])
        m.v12 = Var(m.s1, m.s2)
        m.v124 = Var(m.s1, m.s2, m.s4)
        return m

    def test_no_sets(self):
        m = self.make_model()
        var = m.v12
        sets = (m.s3, m.s4)
        ref_data = {self._hashRef(v) for v in m.v12.values()}

        slices = [slice_ for _, slice_ in slice_component_along_sets(var, sets)]
        self.assertEqual(len(slices), len(ref_data))
        self.assertEqual(len(slices), len(m.s1) * len(m.s2))
        for slice_ in slices:
            self.assertIn(self._hashRef(slice_), ref_data)

    def test_one_set(self):
        m = self.make_model()
        var = m.v124
        sets = (m.s1, m.s3)
        ref_data = {self._hashRef(Reference(m.v124[:, i, j])) for i, j in m.s2 * m.s4}

        slices = [s for _, s in slice_component_along_sets(var, sets)]
        self.assertEqual(len(slices), len(ref_data))
        self.assertEqual(len(slices), len(m.s2) * len(m.s4))
        for slice_ in slices:
            self.assertIn(self._hashRef(Reference(slice_)), ref_data)

    def test_some_sets(self):
        m = self.make_model()
        var = m.v124
        sets = (m.s1, m.s3)
        ref_data = {self._hashRef(Reference(m.v124[:, i, j])) for i, j in m.s2 * m.s4}

        slices = [s for _, s in slice_component_along_sets(var, sets)]
        self.assertEqual(len(slices), len(ref_data))
        self.assertEqual(len(slices), len(m.s2) * len(m.s4))
        for slice_ in slices:
            self.assertIn(self._hashRef(Reference(slice_)), ref_data)

    def test_all_sets(self):
        m = self.make_model()
        var = m.v12
        sets = (m.s1, m.s2)
        ref_data = {self._hashRef(Reference(m.v12[:, :]))}

        slices = [s for _, s in slice_component_along_sets(var, sets)]
        self.assertEqual(len(slices), len(ref_data))
        self.assertEqual(len(slices), 1)
        for slice_ in slices:
            self.assertIn(self._hashRef(Reference(slice_)), ref_data)


class TestExceptional(unittest.TestCase):
    """
    These are the cases that motivate the try/excepts in the slice-checking
    part of the code.
    """

    def test_stop_iteration(self):
        """
        StopIteration is raised if we create an empty slice somewhere
        along the line. It is an open question what we should do in the
        case of an empty slice, but my preference is to omit it so we
        don't return a reference that doesn't admit any valid indices.
        """
        m = ConcreteModel()
        m.s1 = Set(initialize=[1, 2, 3])
        m.s2 = Set(initialize=['a', 'b', 'c'])

        m.v = Var(m.s1, m.s2)

        def con_rule(m, i, j):
            if j == 'a':
                # con[:, 'a'] will be an empty slice
                return Constraint.Skip
            return m.v[i, j] == 5.0

        def vacuous_con_rule(m, i, j):
            # A very odd case
            return Constraint.Skip

        m.con = Constraint(m.s1, m.s2, rule=con_rule)

        with self.assertRaises(StopIteration):
            next(iter(m.con[:, 'a']))

        sets = (m.s1,)
        ctype = Constraint
        sets_list, comps_list = flatten_components_along_sets(m, sets, ctype)
        self.assertEqual(len(comps_list), 1)
        self.assertEqual(len(comps_list[0]), len(m.s2) - 1)

        m.del_component(m.con)
        m.vacuous_con = Constraint(m.s1, m.s2, rule=vacuous_con_rule)

        with self.assertRaises(StopIteration):
            next(iter(m.vacuous_con[...]))

        sets = (m.s1, m.s2)
        ctype = Constraint
        sets_list, comps_list = flatten_components_along_sets(m, sets, ctype)
        self.assertEqual(len(comps_list), 0)

        m.del_component(m.vacuous_con)
        m.del_component(m.v)  # No longer necessary

        # Same behavior can happen for blocks:

        def block_rule(b, i, j):
            b.v = Var()

        m.b = Block(m.s1, m.s2, rule=block_rule)

        for i in m.s1:
            del m.b[i, 'a']

        with self.assertRaises(StopIteration):
            next(iter(m.b[:, 'a'].v))

        sets = (m.s1,)
        ctype = Var
        sets_list, comps_list = flatten_components_along_sets(m, sets, ctype)
        self.assertEqual(len(comps_list), 1)
        self.assertEqual(len(comps_list[0]), len(m.s2) - 1)

        for idx in m.b:
            del m.b[idx]

        with self.assertRaises(StopIteration):
            next(iter(m.b[...].v))

        sets = (m.s1, m.s2)
        ctype = Var
        sets_list, comps_list = flatten_components_along_sets(m, sets, ctype)
        self.assertEqual(len(comps_list), 0)

        # Have a component indexed by all of our sets that doesn't appear
        # when we flatten... seems like somewhat of a contradiction...
        subset_set = ComponentSet(m.b.index_set().subsets())
        for s in sets:
            self.assertIn(s, subset_set)

    def test_descend_stop_iteration(self):
        """
        Even if we construct a non-empty slice, if we provide a bad
        index to descend into, we can end up with no valid blocks
        to descend into. Unclear whether we should raise an error here.
        """
        m = ConcreteModel()
        m.s1 = Set(initialize=[1, 2, 3])
        m.s2 = Set(initialize=['a', 'b'])
        m.v = Var(m.s1, m.s2)

        def b_rule(b, i, j):
            b.v = Var()

        m.b = Block(m.s1, m.s2, rule=b_rule)

        # 'b' will be a bad index to descend into
        for i in m.s1:
            del m.b[i, 'b']

        with self.assertRaises(StopIteration):
            next(iter(m.b[:, 'b']))

        sets = (m.s1, m.s2)
        ctype = Var
        indices = ComponentMap([(m.s2, 'b')])
        sets_list, comps_list = flatten_components_along_sets(
            m, sets, ctype, indices=indices
        )
        for sets, comps in zip(sets_list, comps_list):
            # Here we just check that m.b[:,:].v was not encountered,
            # because of our poor choice of "descend index"
            if len(sets) == 2 and sets[0] is m.s1 and sets[1] is m.s2:
                self.assertEqual(len(comps), 1)
                self.assertEqual(str(ComponentUID(comps[0].referent)), "v[*,*]")
            else:
                raise RuntimeError()

    def test_bad_descend_index(self):
        m = ConcreteModel()
        m.s1 = Set(initialize=[1, 2, 3])
        m.s2 = Set(initialize=['a', 'b'])
        m.v = Var(m.s1, m.s2)

        def b_rule(b, i, j):
            b.v = Var()

        m.b = Block(m.s1, m.s2, rule=b_rule)

        sets = (m.s1, m.s2)
        ctype = Var
        # Here we accidentally provide an index for the wrong set.
        indices = ComponentMap([(m.s1, 'b')])
        # Check that we fail gracefully instead of hitting the StopIteration
        # checked by the above test.
        with self.assertRaisesRegex(ValueError, "bad index"):
            sets_list, comps_list = flatten_components_along_sets(
                m, sets, ctype, indices=indices
            )

    def test_keyerror(self):
        """
        KeyErrors occur when a component that we don't slice
        doesn't have data for some members of its indexing set.
        """
        m = ConcreteModel()
        m.s1 = Set(initialize=[1, 2, 3])
        m.s2 = Set(initialize=['a', 'b', 'c'])

        m.v = Var(m.s1, m.s2)

        def con_rule(m, i, j):
            if j == 'a':
                return Constraint.Skip
            return m.v[i, j] == 5.0

        m.con = Constraint(m.s1, m.s2, rule=con_rule)

        with self.assertRaises(KeyError):
            for idx in m.con.index_set():
                temp = m.con[idx]

        sets = ()
        ctype = Constraint
        sets_list, comps_list = flatten_components_along_sets(m, sets, ctype)
        self.assertEqual(len(sets_list), len(comps_list))
        self.assertEqual(len(sets_list), 1)
        self.assertIs(sets_list[0][0], UnindexedComponent_set)
        self.assertEqual(len(comps_list[0]), len(list(m.con.values())))

        # NOTE: I have been unable to produce this behavior (KeyError)
        # with blocks.


if __name__ == "__main__":
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import pyomo.common.unittest as unittest
from pyomo.core.base import ConcreteModel, Var, Reals
from pyomo.core.beta.dict_objects import (
    VarDict,
    ConstraintDict,
    ObjectiveDict,
    ExpressionDict,
)
from pyomo.core.base.var import VarData
from pyomo.core.base.constraint import ConstraintData
from pyomo.core.base.objective import ObjectiveData
from pyomo.core.base.expression import ExpressionData


class _TestComponentDictBase:
    _ctype = None
    _cdatatype = None

    def setUp(self):
        self.model = ConcreteModel()
        self.model.x = Var()
        # set by derived class
        self._arg = None

    def tearDown(self):
        self.model = None
        self._arg = None

    def test_init1(self):
        model = self.model
        model.c = self._ctype()
        self.assertEqual(model.c.is_indexed(), True)
        self.assertEqual(model.c.is_constructed(), True)

    def test_init2(self):
        model = self.model
        index = ['a', 1, None, (1,), (1, 2)]
        model.c = self._ctype((i, self._cdatatype(self._arg())) for i in index)
        self.assertEqual(model.c.is_indexed(), True)
        self.assertEqual(model.c.is_constructed(), True)
        with self.assertRaises(TypeError):
            model.d = self._ctype(
                *tuple((i, self._cdatatype(self._arg())) for i in index)
            )

    def test_len1(self):
        model = self.model
        model = ConcreteModel()
        model.c = self._ctype()
        self.assertEqual(len(model.c), 0)

    def test_len2(self):
        model = self.model
        index = ['a', 1, None, (1,), (1, 2)]
        model.c = self._ctype((i, self._cdatatype(self._arg())) for i in index)
        self.assertEqual(len(model.c), len(index))

    def test_setitem(self):
        model = self.model
        model = ConcreteModel()
        model.c = self._ctype()
        index = ['a', 1, None, (1,), (1, 2)]
        for i in index:
            self.assertTrue(i not in model.c)
        for cnt, i in enumerate(index, 1):
            model.c[i] = self._cdatatype(self._arg())
            self.assertEqual(len(model.c), cnt)
            self.assertTrue(i in model.c)

    # The immediately following this one was originally written when
    # implicit assignment for update was supported. This should be
    # examined more carefully before supporting it.
    # For now just test that implicit assignment raises an exception
    def test_wrong_type_init(self):
        model = self.model
        index = ['a', 1, None, (1,), (1, 2)]
        with self.assertRaises(TypeError):
            model.c = self._ctype((i, self._arg()) for i in index)

    def test_wrong_type_update(self):
        model = self.model
        index = ['a', 1, None, (1,), (1, 2)]
        model.c = self._ctype()
        with self.assertRaises(TypeError):
            model.c.update((i, self._arg()) for i in index)

    def test_wrong_type_setitem(self):
        model = self.model
        model.c = self._ctype()
        with self.assertRaises(TypeError):
            model.c[1] = self._arg()
        model.c[1] = self._cdatatype(self._arg())
        with self.assertRaises(TypeError):
            model.c[1] = self._arg()

    def test_has_parent_init(self):
        model = self.model
        model.c = self._ctype()
        model.c[1] = self._cdatatype(self._arg())
        with self.assertRaises(ValueError):
            model.d = self._ctype(model.c)
        with self.assertRaises(ValueError):
            model.d = self._ctype(dict(model.c))

    def test_has_parent_update(self):
        model = self.model
        model.c = self._ctype()
        model.c[1] = self._cdatatype(self._arg())
        model.c.update(model.c)
        self.assertEqual(len(model.c), 1)
        model.c.update(dict(model.c))
        self.assertEqual(len(model.c), 1)
        model.d = self._ctype()
        with self.assertRaises(ValueError):
            model.d.update(model.c)
        with self.assertRaises(ValueError):
            model.d.update(dict(model.c))

    def test_has_parent_setitem(self):
        model = self.model
        model.c = self._ctype()
        model.c[1] = self._cdatatype(self._arg())
        model.c[1] = model.c[1]
        with self.assertRaises(ValueError):
            model.c[2] = model.c[1]
        model.d = self._ctype()
        with self.assertRaises(ValueError):
            model.d[None] = model.c[1]

    """
    # make sure an existing Data object is NOT replaced
    # by a call to setitem but simply updated.
    def test_setitem_exists(self):
        model = self.model
        index = ['a', 1, None, (1,), (1,2)]
        model.c = self._ctype((i, self._arg()) for i in index)
        self.assertEqual(len(model.c), len(index))
        for i in index:
            self.assertTrue(i in model.c)
            cdata = model.c[i]
            model.c[i] = self._arg()
            self.assertEqual(len(model.c), len(index))
            self.assertTrue(i in model.c)
            self.assertEqual(id(cdata), id(model.c[i]))
    """

    # make sure an existing Data object IS replaced
    # by a call to setitem and not simply updated.
    def test_setitem_exists_overwrite(self):
        model = self.model
        index = ['a', 1, None, (1,), (1, 2)]
        model.c = self._ctype((i, self._cdatatype(self._arg())) for i in index)
        self.assertEqual(len(model.c), len(index))
        for i in index:
            self.assertTrue(i in model.c)
            cdata = model.c[i]
            model.c[i] = self._cdatatype(self._arg())
            self.assertEqual(len(model.c), len(index))
            self.assertTrue(i in model.c)
            self.assertNotEqual(id(cdata), id(model.c[i]))
            self.assertEqual(cdata.parent_component(), None)

    def test_delitem(self):
        model = self.model
        index = ['a', 1, None, (1,), (1, 2)]
        model.c = self._ctype((i, self._cdatatype(self._arg())) for i in index)
        self.assertEqual(len(model.c), len(index))
        for cnt, i in enumerate(index, 1):
            self.assertTrue(i in model.c)
            cdata = model.c[i]
            self.assertEqual(id(cdata.parent_component()), id(model.c))
            del model.c[i]
            self.assertEqual(len(model.c), len(index) - cnt)
            self.assertTrue(i not in model.c)
            self.assertEqual(cdata.parent_component(), None)

    def test_iter(self):
        model = self.model
        index = ['a', 1, None, (1,), (1, 2)]
        model.c = self._ctype((i, self._cdatatype(self._arg())) for i in index)
        self.assertEqual(len(model.c), len(index))
        comp_index = [i for i in model.c]
        self.assertEqual(len(comp_index), len(index))
        for idx in comp_index:
            self.assertTrue(idx in index)

    def test_model_clone(self):
        model = self.model
        index = ['a', 1, None, (1,), (1, 2)]
        model.c = self._ctype((i, self._cdatatype(self._arg())) for i in index)
        inst = model.clone()
        self.assertNotEqual(id(inst.c), id(model.c))
        for i in index:
            self.assertNotEqual(id(inst.c[i]), id(model.c[i]))

    def test_keys(self):
        model = self.model
        index = ['a', 1, None, (1,), (1, 2)]
        raw_constraint_dict = {i: self._cdatatype(self._arg()) for i in index}
        model.c = self._ctype(raw_constraint_dict)
        self.assertEqual(
            sorted(list(raw_constraint_dict.keys()), key=str),
            sorted(list(model.c.keys()), key=str),
        )

    def test_values(self):
        model = self.model
        index = ['a', 1, None, (1,), (1, 2)]
        raw_constraint_dict = {i: self._cdatatype(self._arg()) for i in index}
        model.c = self._ctype(raw_constraint_dict)
        self.assertEqual(
            sorted(list(id(_v) for _v in raw_constraint_dict.values()), key=str),
            sorted(list(id(_v) for _v in model.c.values()), key=str),
        )

    def test_items(self):
        model = self.model
        index = ['a', 1, None, (1,), (1, 2)]
        raw_constraint_dict = {i: self._cdatatype(self._arg()) for i in index}
        model.c = self._ctype(raw_constraint_dict)
        self.assertEqual(
            sorted(
                list((_i, id(_v)) for _i, _v in raw_constraint_dict.items()), key=str
            ),
            sorted(list((_i, id(_v)) for _i, _v in model.c.items()), key=str),
        )

    def test_update(self):
        model = self.model
        index = ['a', 1, None, (1,), (1, 2)]
        raw_constraint_dict = {i: self._cdatatype(self._arg()) for i in index}
        model.c = self._ctype()
        model.c.update(raw_constraint_dict)
        self.assertEqual(
            sorted(list(raw_constraint_dict.keys()), key=str),
            sorted(list(model.c.keys()), key=str),
        )

    def test_name(self):
        model = self.model
        index = ['a', 1, None, (1,), (1, 2)]
        model.c = self._ctype((i, self._cdatatype(self._arg())) for i in index)
        index_to_string = {}
        index_to_string['a'] = '[a]'
        index_to_string['a,b'] = "['a,b']"
        index_to_string[1] = '[1]'
        index_to_string[None] = '[None]'
        index_to_string[(1,)] = '[(1,)]'
        index_to_string[(1, 2)] = '[1,2]'
        prefix = "c"
        for i in index:
            cdata = model.c[i]
            self.assertEqual(cdata.local_name, cdata.name)
            cname = prefix + index_to_string[i]
            self.assertEqual(cdata.local_name, cname)

    def test_clear(self):
        model = self.model
        model.c = self._ctype()
        model.c[1] = self._cdatatype(self._arg())
        c1 = model.c[1]
        with self.assertRaises(ValueError):
            model.c[None] = c1
        model.d = self._ctype()
        with self.assertRaises(ValueError):
            model.d[1] = c1
        model.c.clear()
        model.d[1] = c1

    def test_eq(self):
        model = self.model
        model.c = self._ctype()
        model.c[1] = self._cdatatype(self._arg())
        model.d = self._ctype()
        model.d[1] = self._cdatatype(self._arg())

        self.assertNotEqual(model.c, [])
        self.assertFalse(model.c == [])

        self.assertTrue(model.c == model.c)
        self.assertEqual(model.c, model.c)
        self.assertTrue(model.c == dict(model.c))
        self.assertEqual(model.c, dict(model.c))
        self.assertTrue(dict(model.c) == model.c)
        self.assertEqual(dict(model.c), model.c)

        self.assertFalse(model.d == model.c)
        self.assertTrue(model.d != model.c)
        self.assertNotEqual(model.d, model.c)

        self.assertFalse(model.c == model.d)
        self.assertTrue(model.c != model.d)
        self.assertNotEqual(model.c, model.d)


class _TestActiveComponentDictBase(_TestComponentDictBase):
    def test_activate(self):
        model = self.model
        index = ['a', 1, None, (1,), (1, 2)]
        model.c = self._ctype((i, self._cdatatype(self._arg())) for i in index)
        self.assertEqual(len(model.c), len(index))
        self.assertEqual(model.c.active, True)
        model.c._active = False
        for i in index:
            model.c[i]._active = False
        self.assertEqual(model.c.active, False)
        for i in index:
            self.assertEqual(model.c[i].active, False)
        model.c.activate()
        self.assertEqual(model.c.active, True)

    def test_activate(self):
        model = self.model
        index = ['a', 1, None, (1,), (1, 2)]
        model.c = self._ctype((i, self._cdatatype(self._arg())) for i in index)
        self.assertEqual(len(model.c), len(index))
        self.assertEqual(model.c.active, True)
        for i in index:
            self.assertEqual(model.c[i].active, True)
        model.c.deactivate()
        self.assertEqual(model.c.active, False)
        for i in index:
            self.assertEqual(model.c[i].active, False)

    # Befault, assigning a new component to a dict container makes it
    # active (unless the default active state for the container
    # datatype is False, which is not the case for any currently
    # existing implementations).
    def test_active(self):
        model = self.model
        model.c = self._ctype()
        self.assertEqual(model.c.active, True)
        model.c.deactivate()
        self.assertEqual(model.c.active, False)
        model.c[1] = self._cdatatype(self._arg())
        self.assertEqual(model.c.active, True)


class TestVarDict(_TestComponentDictBase, unittest.TestCase):
    # Note: the updated VarData class only takes an optional
    # parent argument (you no longer pass the domain in)
    _ctype = VarDict
    _cdatatype = lambda self, arg: VarData()

    def setUp(self):
        _TestComponentDictBase.setUp(self)
        self._arg = lambda: Reals


class TestExpressionDict(_TestComponentDictBase, unittest.TestCase):
    _ctype = ExpressionDict
    _cdatatype = ExpressionData

    def setUp(self):
        _TestComponentDictBase.setUp(self)
        self._arg = lambda: self.model.x**3


#
# Test components that include activate/deactivate
# functionality.
#


class TestConstraintDict(_TestActiveComponentDictBase, unittest.TestCase):
    _ctype = ConstraintDict
    _cdatatype = ConstraintData

    def setUp(self):
        _TestComponentDictBase.setUp(self)
        self._arg = lambda: self.model.x >= 1


class TestObjectiveDict(_TestActiveComponentDictBase, unittest.TestCase):
    _ctype = ObjectiveDict
    _cdatatype = ObjectiveData

    def setUp(self):
        _TestComponentDictBase.setUp(self)
        self._arg = lambda: self.model.x**2


if __name__ == "__main__":
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import abc
import enum
import os
import re
import weakref

from typing import (
    TYPE_CHECKING,
    Sequence,
    Dict,
    Optional,
    Mapping,
    NoReturn,
    List,
    Tuple,
    MutableMapping,
)

from pyomo.common.config import ConfigDict, ConfigValue, NonNegativeFloat
from pyomo.common.errors import ApplicationError
from pyomo.common.enums import IntEnum, SolverAPIVersion
from pyomo.common.factory import Factory
from pyomo.common.timing import HierarchicalTimer
from pyomo.core.base.constraint import ConstraintData, Constraint
from pyomo.core.base.sos import SOSConstraintData, SOSConstraint
from pyomo.core.base.var import VarData, Var
from pyomo.core.base.param import ParamData, Param
from pyomo.core.base.block import BlockData, Block
from pyomo.core.base.objective import ObjectiveData
from pyomo.common.collections import ComponentMap
from .utils.get_objective import get_objective
from .utils.collect_vars_and_named_exprs import collect_vars_and_named_exprs
from pyomo.opt.base import SolverFactory as LegacySolverFactory
from pyomo.opt.results.results_ import SolverResults as LegacySolverResults
from pyomo.opt.results.solution import (
    Solution as LegacySolution,
    SolutionStatus as LegacySolutionStatus,
)
from pyomo.opt.results.solver import (
    TerminationCondition as LegacyTerminationCondition,
    SolverStatus as LegacySolverStatus,
)
from pyomo.core.kernel.objective import minimize
from pyomo.core.base import SymbolMap
from .cmodel import cmodel, cmodel_available
from pyomo.core.staleflag import StaleFlagManager
from pyomo.core.expr.numvalue import NumericConstant


class TerminationCondition(enum.Enum):
    """
    An enumeration for checking the termination condition of solvers
    """

    unknown = 0
    """unknown serves as both a default value, and it is used when no other enum member makes sense"""

    maxTimeLimit = 1
    """The solver exited due to a time limit"""

    maxIterations = 2
    """The solver exited due to an iteration limit """

    objectiveLimit = 3
    """The solver exited due to an objective limit"""

    minStepLength = 4
    """The solver exited due to a minimum step length"""

    optimal = 5
    """The solver exited with the optimal solution"""

    unbounded = 8
    """The solver exited because the problem is unbounded"""

    infeasible = 9
    """The solver exited because the problem is infeasible"""

    infeasibleOrUnbounded = 10
    """The solver exited because the problem is either infeasible or unbounded"""

    error = 11
    """The solver exited due to an error"""

    interrupted = 12
    """The solver exited because it was interrupted"""

    licensingProblems = 13
    """The solver exited due to licensing problems"""


class SolverConfig(ConfigDict):
    """
    Common configuration options for all APPSI solver interfaces


    Attributes
    ----------
    time_limit: float
        Time limit for the solver
    stream_solver: bool
        If True, then the solver log goes to stdout
    load_solution: bool
        If False, then the values of the primal variables will not be
        loaded into the model
    symbolic_solver_labels: bool
        If True, the names given to the solver will reflect the names
        of the pyomo components. Cannot be changed after set_instance
        is called.
    report_timing: bool
        If True, then some timing information will be printed at the
        end of the solve.
    """

    def __init__(
        self,
        description=None,
        doc=None,
        implicit=False,
        implicit_domain=None,
        visibility=0,
    ):
        super(SolverConfig, self).__init__(
            description=description,
            doc=doc,
            implicit=implicit,
            implicit_domain=implicit_domain,
            visibility=visibility,
        )

        self.declare('time_limit', ConfigValue(domain=NonNegativeFloat))
        self.declare('warmstart', ConfigValue(domain=bool))
        self.declare('stream_solver', ConfigValue(domain=bool))
        self.declare('load_solution', ConfigValue(domain=bool))
        self.declare('symbolic_solver_labels', ConfigValue(domain=bool))
        self.declare('report_timing', ConfigValue(domain=bool))

        self.time_limit: Optional[float] = None
        self.warmstart: bool = False
        self.stream_solver: bool = False
        self.load_solution: bool = True
        self.symbolic_solver_labels: bool = False
        self.report_timing: bool = False


class MIPSolverConfig(SolverConfig):
    """
    Configuration options common to all MIP solvers

    Attributes
    ----------
    mip_gap: float
        Solver will terminate if the mip gap is less than mip_gap
    relax_integrality: bool
        If True, all integer variables will be relaxed to continuous
        variables before solving
    """

    def __init__(
        self,
        description=None,
        doc=None,
        implicit=False,
        implicit_domain=None,
        visibility=0,
    ):
        super(MIPSolverConfig, self).__init__(
            description=description,
            doc=doc,
            implicit=implicit,
            implicit_domain=implicit_domain,
            visibility=visibility,
        )

        self.declare('mip_gap', ConfigValue(domain=NonNegativeFloat))
        self.declare('relax_integrality', ConfigValue(domain=bool))

        self.mip_gap: Optional[float] = None
        self.relax_integrality: bool = False


class SolutionLoaderBase(abc.ABC):
    def load_vars(self, vars_to_load: Optional[Sequence[VarData]] = None) -> NoReturn:
        """
        Load the solution of the primal variables into the value attribute of the variables.

        Parameters
        ----------
        vars_to_load: list
            A list of the variables whose solution should be loaded. If vars_to_load is None, then the solution
            to all primal variables will be loaded.
        """
        for v, val in self.get_primals(vars_to_load=vars_to_load).items():
            v.set_value(val, skip_validation=True)
        StaleFlagManager.mark_all_as_stale(delayed=True)

    @abc.abstractmethod
    def get_primals(
        self, vars_to_load: Optional[Sequence[VarData]] = None
    ) -> Mapping[VarData, float]:
        """
        Returns a ComponentMap mapping variable to var value.

        Parameters
        ----------
        vars_to_load: list
            A list of the variables whose solution value should be retrieved. If vars_to_load is None,
            then the values for all variables will be retrieved.

        Returns
        -------
        primals: ComponentMap
            Maps variables to solution values
        """
        pass

    def get_duals(
        self, cons_to_load: Optional[Sequence[ConstraintData]] = None
    ) -> Dict[ConstraintData, float]:
        """
        Returns a dictionary mapping constraint to dual value.

        Parameters
        ----------
        cons_to_load: list
            A list of the constraints whose duals should be retrieved. If cons_to_load is None, then the duals for all
            constraints will be retrieved.

        Returns
        -------
        duals: dict
            Maps constraints to dual values
        """
        raise NotImplementedError(f'{type(self)} does not support the get_duals method')

    def get_slacks(
        self, cons_to_load: Optional[Sequence[ConstraintData]] = None
    ) -> Dict[ConstraintData, float]:
        """
        Returns a dictionary mapping constraint to slack.

        Parameters
        ----------
        cons_to_load: list
            A list of the constraints whose duals should be loaded. If cons_to_load is None, then the duals for all
            constraints will be loaded.

        Returns
        -------
        slacks: dict
            Maps constraints to slacks
        """
        raise NotImplementedError(
            f'{type(self)} does not support the get_slacks method'
        )

    def get_reduced_costs(
        self, vars_to_load: Optional[Sequence[VarData]] = None
    ) -> Mapping[VarData, float]:
        """
        Returns a ComponentMap mapping variable to reduced cost.

        Parameters
        ----------
        vars_to_load: list
            A list of the variables whose reduced cost should be retrieved. If vars_to_load is None, then the
            reduced costs for all variables will be loaded.

        Returns
        -------
        reduced_costs: ComponentMap
            Maps variables to reduced costs
        """
        raise NotImplementedError(
            f'{type(self)} does not support the get_reduced_costs method'
        )


class SolutionLoader(SolutionLoaderBase):
    def __init__(
        self,
        primals: Optional[MutableMapping],
        duals: Optional[MutableMapping],
        slacks: Optional[MutableMapping],
        reduced_costs: Optional[MutableMapping],
    ):
        """
        Parameters
        ----------
        primals: dict
            maps id(Var) to (var, value)
        duals: dict
            maps Constraint to dual value
        slacks: dict
            maps Constraint to slack value
        reduced_costs: dict
            maps id(Var) to (var, reduced_cost)
        """
        self._primals = primals
        self._duals = duals
        self._slacks = slacks
        self._reduced_costs = reduced_costs

    def get_primals(
        self, vars_to_load: Optional[Sequence[VarData]] = None
    ) -> Mapping[VarData, float]:
        if self._primals is None:
            raise RuntimeError(
                'Solution loader does not currently have a valid solution. Please '
                'check the termination condition.'
            )
        if vars_to_load is None:
            return ComponentMap(self._primals.values())
        else:
            primals = ComponentMap()
            for v in vars_to_load:
                primals[v] = self._primals[id(v)][1]
            return primals

    def get_duals(
        self, cons_to_load: Optional[Sequence[ConstraintData]] = None
    ) -> Dict[ConstraintData, float]:
        if self._duals is None:
            raise RuntimeError(
                'Solution loader does not currently have valid duals. Please '
                'check the termination condition and ensure the solver returns duals '
                'for the given problem type.'
            )
        if cons_to_load is None:
            duals = dict(self._duals)
        else:
            duals = dict()
            for c in cons_to_load:
                duals[c] = self._duals[c]
        return duals

    def get_slacks(
        self, cons_to_load: Optional[Sequence[ConstraintData]] = None
    ) -> Dict[ConstraintData, float]:
        if self._slacks is None:
            raise RuntimeError(
                'Solution loader does not currently have valid slacks. Please '
                'check the termination condition and ensure the solver returns slacks '
                'for the given problem type.'
            )
        if cons_to_load is None:
            slacks = dict(self._slacks)
        else:
            slacks = dict()
            for c in cons_to_load:
                slacks[c] = self._slacks[c]
        return slacks

    def get_reduced_costs(
        self, vars_to_load: Optional[Sequence[VarData]] = None
    ) -> Mapping[VarData, float]:
        if self._reduced_costs is None:
            raise RuntimeError(
                'Solution loader does not currently have valid reduced costs. Please '
                'check the termination condition and ensure the solver returns reduced '
                'costs for the given problem type.'
            )
        if vars_to_load is None:
            rc = ComponentMap(self._reduced_costs.values())
        else:
            rc = ComponentMap()
            for v in vars_to_load:
                rc[v] = self._reduced_costs[id(v)][1]
        return rc


class Results:
    """
    Base class for all APPSI solver results

    Attributes
    ----------
    termination_condition: TerminationCondition
        The reason the solver exited. This is a member of the
        TerminationCondition enum.
    best_feasible_objective: float
        If a feasible solution was found, this is the objective value of
        the best solution found. If no feasible solution was found, this is
        None.
    best_objective_bound: float
        The best objective bound found. For minimization problems, this is
        the lower bound. For maximization problems, this is the upper bound.
        For solvers that do not provide an objective bound, this should be -inf
        (minimization) or inf (maximization)

    Example
    -------
    Here is an example workflow:

        >>> import pyomo.environ as pyo
        >>> from pyomo.contrib import appsi
        >>> m = pyo.ConcreteModel()
        >>> m.x = pyo.Var()
        >>> m.obj = pyo.Objective(expr=m.x**2)
        >>> opt = appsi.solvers.Ipopt()
        >>> opt.config.load_solution = False
        >>> results = opt.solve(m) #doctest:+SKIP
        >>> if results.termination_condition == appsi.base.TerminationCondition.optimal: #doctest:+SKIP
        ...     print('optimal solution found: ', results.best_feasible_objective) #doctest:+SKIP
        ...     results.solution_loader.load_vars() #doctest:+SKIP
        ...     print('the optimal value of x is ', m.x.value) #doctest:+SKIP
        ... elif results.best_feasible_objective is not None: #doctest:+SKIP
        ...     print('sub-optimal but feasible solution found: ', results.best_feasible_objective) #doctest:+SKIP
        ...     results.solution_loader.load_vars(vars_to_load=[m.x]) #doctest:+SKIP
        ...     print('The value of x in the feasible solution is ', m.x.value) #doctest:+SKIP
        ... elif results.termination_condition in {appsi.base.TerminationCondition.maxIterations, appsi.base.TerminationCondition.maxTimeLimit}: #doctest:+SKIP
        ...     print('No feasible solution was found. The best lower bound found was ', results.best_objective_bound) #doctest:+SKIP
        ... else: #doctest:+SKIP
        ...     print('The following termination condition was encountered: ', results.termination_condition) #doctest:+SKIP
    """

    def __init__(self):
        self.solution_loader: SolutionLoaderBase = SolutionLoader(
            None, None, None, None
        )
        self.termination_condition: TerminationCondition = TerminationCondition.unknown
        self.best_feasible_objective: Optional[float] = None
        self.best_objective_bound: Optional[float] = None

    def __str__(self):
        s = ''
        s += 'termination_condition: ' + str(self.termination_condition) + '\n'
        s += 'best_feasible_objective: ' + str(self.best_feasible_objective) + '\n'
        s += 'best_objective_bound: ' + str(self.best_objective_bound)
        return s


class UpdateConfig(ConfigDict):
    """
    Config options common to all persistent solvers

    Attributes
    ----------
    check_for_new_or_removed_constraints: bool
    check_for_new_or_removed_vars: bool
    check_for_new_or_removed_params: bool
    update_constraints: bool
    update_vars: bool
    update_params: bool
    update_named_expressions: bool
    """

    def __init__(
        self,
        description=None,
        doc=None,
        implicit=False,
        implicit_domain=None,
        visibility=0,
    ):
        if doc is None:
            doc = 'Configuration options to detect changes in model between solves'
        super(UpdateConfig, self).__init__(
            description=description,
            doc=doc,
            implicit=implicit,
            implicit_domain=implicit_domain,
            visibility=visibility,
        )

        self.declare(
            'check_for_new_or_removed_constraints',
            ConfigValue(
                domain=bool,
                default=True,
                doc="""
                If False, new/old constraints will not be automatically detected on subsequent
                solves. Use False only when manually updating the solver with opt.add_constraints()
                and opt.remove_constraints() or when you are certain constraints are not being
                added to/removed from the model.""",
            ),
        )
        self.declare(
            'check_for_new_or_removed_vars',
            ConfigValue(
                domain=bool,
                default=True,
                doc="""
                If False, new/old variables will not be automatically detected on subsequent 
                solves. Use False only when manually updating the solver with opt.add_variables() and 
                opt.remove_variables() or when you are certain variables are not being added to /
                removed from the model.""",
            ),
        )
        self.declare(
            'check_for_new_or_removed_params',
            ConfigValue(
                domain=bool,
                default=True,
                doc="""
                If False, new/old parameters will not be automatically detected on subsequent 
                solves. Use False only when manually updating the solver with opt.add_params() and 
                opt.remove_params() or when you are certain parameters are not being added to /
                removed from the model.""",
            ),
        )
        self.declare(
            'check_for_new_objective',
            ConfigValue(
                domain=bool,
                default=True,
                doc="""
                If False, new/old objectives will not be automatically detected on subsequent 
                solves. Use False only when manually updating the solver with opt.set_objective() or 
                when you are certain objectives are not being added to / removed from the model.""",
            ),
        )
        self.declare(
            'update_constraints',
            ConfigValue(
                domain=bool,
                default=True,
                doc="""
                If False, changes to existing constraints will not be automatically detected on 
                subsequent solves. This includes changes to the lower, body, and upper attributes of 
                constraints. Use False only when manually updating the solver with 
                opt.remove_constraints() and opt.add_constraints() or when you are certain constraints 
                are not being modified.""",
            ),
        )
        self.declare(
            'update_vars',
            ConfigValue(
                domain=bool,
                default=True,
                doc="""
                If False, changes to existing variables will not be automatically detected on 
                subsequent solves. This includes changes to the lb, ub, domain, and fixed 
                attributes of variables. Use False only when manually updating the solver with 
                opt.update_variables() or when you are certain variables are not being modified.""",
            ),
        )
        self.declare(
            'update_params',
            ConfigValue(
                domain=bool,
                default=True,
                doc="""
                If False, changes to parameter values will not be automatically detected on 
                subsequent solves. Use False only when manually updating the solver with 
                opt.update_params() or when you are certain parameters are not being modified.""",
            ),
        )
        self.declare(
            'update_named_expressions',
            ConfigValue(
                domain=bool,
                default=True,
                doc="""
                If False, changes to Expressions will not be automatically detected on 
                subsequent solves. Use False only when manually updating the solver with 
                opt.remove_constraints() and opt.add_constraints() or when you are certain 
                Expressions are not being modified.""",
            ),
        )
        self.declare(
            'update_objective',
            ConfigValue(
                domain=bool,
                default=True,
                doc="""
                If False, changes to objectives will not be automatically detected on 
                subsequent solves. This includes the expr and sense attributes of objectives. Use 
                False only when manually updating the solver with opt.set_objective() or when you are 
                certain objectives are not being modified.""",
            ),
        )
        self.declare(
            'treat_fixed_vars_as_params',
            ConfigValue(
                domain=bool,
                default=True,
                doc="""
                This is an advanced option that should only be used in special circumstances. 
                With the default setting of True, fixed variables will be treated like parameters. 
                This means that z == x*y will be linear if x or y is fixed and the constraint 
                can be written to an LP file. If the value of the fixed variable gets changed, we have 
                to completely reprocess all constraints using that variable. If 
                treat_fixed_vars_as_params is False, then constraints will be processed as if fixed 
                variables are not fixed, and the solver will be told the variable is fixed. This means 
                z == x*y could not be written to an LP file even if x and/or y is fixed. However, 
                updating the values of fixed variables is much faster this way.""",
            ),
        )

        self.check_for_new_or_removed_constraints: bool = True
        self.check_for_new_or_removed_vars: bool = True
        self.check_for_new_or_removed_params: bool = True
        self.check_for_new_objective: bool = True
        self.update_constraints: bool = True
        self.update_vars: bool = True
        self.update_params: bool = True
        self.update_named_expressions: bool = True
        self.update_objective: bool = True
        self.treat_fixed_vars_as_params: bool = True


class Solver(abc.ABC):
    class Availability(IntEnum):
        NotFound = 0
        BadVersion = -1
        BadLicense = -2
        FullLicense = 1
        LimitedLicense = 2
        NeedsCompiledExtension = -3

        def __bool__(self):
            return self._value_ > 0

        def __format__(self, format_spec):
            # We want general formatting of this Enum to return the
            # formatted string value and not the int (which is the
            # default implementation from IntEnum)
            return format(self.name, format_spec)

        def __str__(self):
            # Note: Python 3.11 changed the core enums so that the
            # "mixin" type for standard enums overrides the behavior
            # specified in __format__.  We will override str() here to
            # preserve the previous behavior
            return self.name

    @classmethod
    def api_version(self):
        """
        Return the public API supported by this interface.

        Returns
        -------
        ~pyomo.common.enums.SolverAPIVersion
            A solver API enum object
        """
        return SolverAPIVersion.APPSI

    @abc.abstractmethod
    def solve(self, model: BlockData, timer: HierarchicalTimer = None) -> Results:
        """
        Solve a Pyomo model.

        Parameters
        ----------
        model: BlockData
            The Pyomo model to be solved
        timer: HierarchicalTimer
            An option timer for reporting timing

        Returns
        -------
        results: ~pyomo.contrib.appsi.base.Results
            A results object
        """
        pass

    @abc.abstractmethod
    def available(self):
        """Test if the solver is available on this system.

        Nominally, this will return True if the solver interface is
        valid and can be used to solve problems and False if it cannot.

        Note that for licensed solvers there are a number of "levels" of
        available: depending on the license, the solver may be available
        with limitations on problem size or runtime (e.g., 'demo'
        vs. 'community' vs. 'full').  In these cases, the solver may
        return a subclass of enum.IntEnum, with members that resolve to
        True if the solver is available (possibly with limitations).
        The Enum may also have multiple members that all resolve to
        False indicating the reason why the interface is not available
        (not found, bad license, unsupported version, etc).

        Returns
        -------
        available: Solver.Availability
            An enum that indicates "how available" the solver is.
            Note that the enum can be cast to bool, which will
            be True if the solver is runnable at all and False
            otherwise.
        """
        pass

    @abc.abstractmethod
    def version(self) -> Tuple:
        """
        Returns
        -------
        version: tuple
            A tuple representing the version
        """

    @property
    @abc.abstractmethod
    def config(self):
        """
        An object for configuring solve options.

        Returns
        -------
        ~pyomo.contrib.appsi.base.SolverConfig
            An object for configuring pyomo solve options such as the time limit.
            These options are mostly independent of the solver.
        """
        pass

    @property
    @abc.abstractmethod
    def symbol_map(self):
        pass

    def is_persistent(self):
        """
        Returns
        -------
        is_persistent: bool
            True if the solver is a persistent solver.
        """
        return False


class PersistentSolver(Solver):
    def is_persistent(self):
        return True

    def load_vars(self, vars_to_load: Optional[Sequence[VarData]] = None) -> NoReturn:
        """
        Load the solution of the primal variables into the value attribute of the variables.

        Parameters
        ----------
        vars_to_load: list
            A list of the variables whose solution should be loaded. If vars_to_load is None, then the solution
            to all primal variables will be loaded.
        """
        for v, val in self.get_primals(vars_to_load=vars_to_load).items():
            v.set_value(val, skip_validation=True)
        StaleFlagManager.mark_all_as_stale(delayed=True)

    @abc.abstractmethod
    def get_primals(
        self, vars_to_load: Optional[Sequence[VarData]] = None
    ) -> Mapping[VarData, float]:
        pass

    def get_duals(
        self, cons_to_load: Optional[Sequence[ConstraintData]] = None
    ) -> Dict[ConstraintData, float]:
        """
        Declare sign convention in docstring here.

        Parameters
        ----------
        cons_to_load: list
            A list of the constraints whose duals should be loaded. If cons_to_load is None, then the duals for all
            constraints will be loaded.

        Returns
        -------
        duals: dict
            Maps constraints to dual values
        """
        raise NotImplementedError(
            '{0} does not support the get_duals method'.format(type(self))
        )

    def get_slacks(
        self, cons_to_load: Optional[Sequence[ConstraintData]] = None
    ) -> Dict[ConstraintData, float]:
        """
        Parameters
        ----------
        cons_to_load: list
            A list of the constraints whose slacks should be loaded. If cons_to_load is None, then the slacks for all
            constraints will be loaded.

        Returns
        -------
        slacks: dict
            Maps constraints to slack values
        """
        raise NotImplementedError(
            '{0} does not support the get_slacks method'.format(type(self))
        )

    def get_reduced_costs(
        self, vars_to_load: Optional[Sequence[VarData]] = None
    ) -> Mapping[VarData, float]:
        """
        Parameters
        ----------
        vars_to_load: list
            A list of the variables whose reduced cost should be loaded. If vars_to_load is None, then all reduced costs
            will be loaded.

        Returns
        -------
        reduced_costs: ComponentMap
            Maps variable to reduced cost
        """
        raise NotImplementedError(
            '{0} does not support the get_reduced_costs method'.format(type(self))
        )

    @property
    @abc.abstractmethod
    def update_config(self) -> UpdateConfig:
        pass

    @abc.abstractmethod
    def set_instance(self, model):
        pass

    @abc.abstractmethod
    def add_variables(self, variables: List[VarData]):
        pass

    @abc.abstractmethod
    def add_params(self, params: List[ParamData]):
        pass

    @abc.abstractmethod
    def add_constraints(self, cons: List[ConstraintData]):
        pass

    @abc.abstractmethod
    def add_block(self, block: BlockData):
        pass

    @abc.abstractmethod
    def remove_variables(self, variables: List[VarData]):
        pass

    @abc.abstractmethod
    def remove_params(self, params: List[ParamData]):
        pass

    @abc.abstractmethod
    def remove_constraints(self, cons: List[ConstraintData]):
        pass

    @abc.abstractmethod
    def remove_block(self, block: BlockData):
        pass

    @abc.abstractmethod
    def set_objective(self, obj: ObjectiveData):
        pass

    @abc.abstractmethod
    def update_variables(self, variables: List[VarData]):
        pass

    @abc.abstractmethod
    def update_params(self):
        pass


class PersistentSolutionLoader(SolutionLoaderBase):
    def __init__(self, solver: PersistentSolver):
        self._solver = solver
        self._valid = True

    def _assert_solution_still_valid(self):
        if not self._valid:
            raise RuntimeError('The results in the solver are no longer valid.')

    def get_primals(self, vars_to_load=None):
        self._assert_solution_still_valid()
        return self._solver.get_primals(vars_to_load=vars_to_load)

    def get_duals(
        self, cons_to_load: Optional[Sequence[ConstraintData]] = None
    ) -> Dict[ConstraintData, float]:
        self._assert_solution_still_valid()
        return self._solver.get_duals(cons_to_load=cons_to_load)

    def get_slacks(
        self, cons_to_load: Optional[Sequence[ConstraintData]] = None
    ) -> Dict[ConstraintData, float]:
        self._assert_solution_still_valid()
        return self._solver.get_slacks(cons_to_load=cons_to_load)

    def get_reduced_costs(
        self, vars_to_load: Optional[Sequence[VarData]] = None
    ) -> Mapping[VarData, float]:
        self._assert_solution_still_valid()
        return self._solver.get_reduced_costs(vars_to_load=vars_to_load)

    def invalidate(self):
        self._valid = False


"""
What can change in a pyomo model?
- variables added or removed
- constraints added or removed
- objective changed
- objective expr changed
- params added or removed
- variable modified
  - lb
  - ub
  - fixed or unfixed
  - domain
  - value
- constraint modified
  - lower
  - upper
  - body
  - active or not
- named expressions modified
  - expr
- param modified
  - value

Ideas:
- Consider explicitly handling deactivated constraints; favor deactivation over removal
  and activation over addition
  
Notes:
- variable bounds cannot be updated with mutable params; you must call update_variables
"""


class PersistentBase(abc.ABC):
    def __init__(self, only_child_vars=False):
        self._model = None
        self._active_constraints = dict()  # maps constraint to (lower, body, upper)
        self._vars = dict()  # maps var id to (var, lb, ub, fixed, domain, value)
        self._params = dict()  # maps param id to param
        self._objective = None
        self._objective_expr = None
        self._objective_sense = None
        self._named_expressions = (
            dict()
        )  # maps constraint to list of tuples (named_expr, named_expr.expr)
        self._external_functions = ComponentMap()
        self._obj_named_expressions = list()
        self._update_config = UpdateConfig()
        self._referenced_variables = (
            dict()
        )  # var_id: [dict[constraints, None], dict[sos constraints, None], None or objective]
        self._vars_referenced_by_con = dict()
        self._vars_referenced_by_obj = list()
        self._expr_types = None
        self.use_extensions = False
        self._only_child_vars = only_child_vars

    @property
    def update_config(self):
        return self._update_config

    @update_config.setter
    def update_config(self, val: UpdateConfig):
        self._update_config = val

    def set_instance(self, model):
        saved_update_config = self.update_config
        self.__init__(only_child_vars=self._only_child_vars)
        self.update_config = saved_update_config
        self._model = model
        if self.use_extensions and cmodel_available:
            self._expr_types = cmodel.PyomoExprTypes()
        self.add_block(model)
        if self._objective is None:
            self.set_objective(None)

    @abc.abstractmethod
    def _add_variables(self, variables: List[VarData]):
        pass

    def add_variables(self, variables: List[VarData]):
        for v in variables:
            if id(v) in self._referenced_variables:
                raise ValueError(
                    'variable {name} has already been added'.format(name=v.name)
                )
            self._referenced_variables[id(v)] = [dict(), dict(), None]
            self._vars[id(v)] = (
                v,
                v._lb,
                v._ub,
                v.fixed,
                v.domain.get_interval(),
                v.value,
            )
        self._add_variables(variables)

    @abc.abstractmethod
    def _add_params(self, params: List[ParamData]):
        pass

    def add_params(self, params: List[ParamData]):
        for p in params:
            self._params[id(p)] = p
        self._add_params(params)

    @abc.abstractmethod
    def _add_constraints(self, cons: List[ConstraintData]):
        pass

    def _check_for_new_vars(self, variables: List[VarData]):
        new_vars = dict()
        for v in variables:
            v_id = id(v)
            if v_id not in self._referenced_variables:
                new_vars[v_id] = v
        self.add_variables(list(new_vars.values()))

    def _check_to_remove_vars(self, variables: List[VarData]):
        vars_to_remove = dict()
        for v in variables:
            v_id = id(v)
            ref_cons, ref_sos, ref_obj = self._referenced_variables[v_id]
            if len(ref_cons) == 0 and len(ref_sos) == 0 and ref_obj is None:
                vars_to_remove[v_id] = v
        self.remove_variables(list(vars_to_remove.values()))

    def add_constraints(self, cons: List[ConstraintData]):
        all_fixed_vars = dict()
        for con in cons:
            if con in self._named_expressions:
                raise ValueError(
                    'constraint {name} has already been added'.format(name=con.name)
                )
            self._active_constraints[con] = con.expr
            if self.use_extensions and cmodel_available:
                tmp = cmodel.prep_for_repn(con.body, self._expr_types)
            else:
                tmp = collect_vars_and_named_exprs(con.body)
            named_exprs, variables, fixed_vars, external_functions = tmp
            if not self._only_child_vars:
                self._check_for_new_vars(variables)
            self._named_expressions[con] = [(e, e.expr) for e in named_exprs]
            if len(external_functions) > 0:
                self._external_functions[con] = external_functions
            self._vars_referenced_by_con[con] = variables
            for v in variables:
                self._referenced_variables[id(v)][0][con] = None
            if not self.update_config.treat_fixed_vars_as_params:
                for v in fixed_vars:
                    v.unfix()
                    all_fixed_vars[id(v)] = v
        self._add_constraints(cons)
        for v in all_fixed_vars.values():
            v.fix()

    @abc.abstractmethod
    def _add_sos_constraints(self, cons: List[SOSConstraintData]):
        pass

    def add_sos_constraints(self, cons: List[SOSConstraintData]):
        for con in cons:
            if con in self._vars_referenced_by_con:
                raise ValueError(
                    'constraint {name} has already been added'.format(name=con.name)
                )
            self._active_constraints[con] = tuple()
            variables = con.get_variables()
            if not self._only_child_vars:
                self._check_for_new_vars(variables)
            self._named_expressions[con] = list()
            self._vars_referenced_by_con[con] = variables
            for v in variables:
                self._referenced_variables[id(v)][1][con] = None
        self._add_sos_constraints(cons)

    @abc.abstractmethod
    def _set_objective(self, obj: ObjectiveData):
        pass

    def set_objective(self, obj: ObjectiveData):
        if self._objective is not None:
            for v in self._vars_referenced_by_obj:
                self._referenced_variables[id(v)][2] = None
            if not self._only_child_vars:
                self._check_to_remove_vars(self._vars_referenced_by_obj)
            self._external_functions.pop(self._objective, None)
        if obj is not None:
            self._objective = obj
            self._objective_expr = obj.expr
            self._objective_sense = obj.sense
            if self.use_extensions and cmodel_available:
                tmp = cmodel.prep_for_repn(obj.expr, self._expr_types)
            else:
                tmp = collect_vars_and_named_exprs(obj.expr)
            named_exprs, variables, fixed_vars, external_functions = tmp
            if not self._only_child_vars:
                self._check_for_new_vars(variables)
            self._obj_named_expressions = [(i, i.expr) for i in named_exprs]
            if len(external_functions) > 0:
                self._external_functions[obj] = external_functions
            self._vars_referenced_by_obj = variables
            for v in variables:
                self._referenced_variables[id(v)][2] = obj
            if not self.update_config.treat_fixed_vars_as_params:
                for v in fixed_vars:
                    v.unfix()
            self._set_objective(obj)
            for v in fixed_vars:
                v.fix()
        else:
            self._vars_referenced_by_obj = list()
            self._objective = None
            self._objective_expr = None
            self._objective_sense = None
            self._obj_named_expressions = list()
            self._set_objective(obj)

    def add_block(self, block):
        param_dict = dict()
        for p in block.component_objects(Param, descend_into=True):
            if p.mutable:
                for _p in p.values():
                    param_dict[id(_p)] = _p
        self.add_params(list(param_dict.values()))
        if self._only_child_vars:
            self.add_variables(
                list(
                    dict(
                        (id(var), var)
                        for var in block.component_data_objects(Var, descend_into=True)
                    ).values()
                )
            )
        self.add_constraints(
            [
                con
                for con in block.component_data_objects(
                    Constraint, descend_into=True, active=True
                )
            ]
        )
        self.add_sos_constraints(
            [
                con
                for con in block.component_data_objects(
                    SOSConstraint, descend_into=True, active=True
                )
            ]
        )
        obj = get_objective(block)
        if obj is not None:
            self.set_objective(obj)

    @abc.abstractmethod
    def _remove_constraints(self, cons: List[ConstraintData]):
        pass

    def remove_constraints(self, cons: List[ConstraintData]):
        self._remove_constraints(cons)
        for con in cons:
            if con not in self._named_expressions:
                raise ValueError(
                    'cannot remove constraint {name} - it was not added'.format(
                        name=con.name
                    )
                )
            for v in self._vars_referenced_by_con[con]:
                self._referenced_variables[id(v)][0].pop(con)
            if not self._only_child_vars:
                self._check_to_remove_vars(self._vars_referenced_by_con[con])
            del self._active_constraints[con]
            del self._named_expressions[con]
            self._external_functions.pop(con, None)
            del self._vars_referenced_by_con[con]

    @abc.abstractmethod
    def _remove_sos_constraints(self, cons: List[SOSConstraintData]):
        pass

    def remove_sos_constraints(self, cons: List[SOSConstraintData]):
        self._remove_sos_constraints(cons)
        for con in cons:
            if con not in self._vars_referenced_by_con:
                raise ValueError(
                    'cannot remove constraint {name} - it was not added'.format(
                        name=con.name
                    )
                )
            for v in self._vars_referenced_by_con[con]:
                self._referenced_variables[id(v)][1].pop(con)
            self._check_to_remove_vars(self._vars_referenced_by_con[con])
            del self._active_constraints[con]
            del self._named_expressions[con]
            del self._vars_referenced_by_con[con]

    @abc.abstractmethod
    def _remove_variables(self, variables: List[VarData]):
        pass

    def remove_variables(self, variables: List[VarData]):
        self._remove_variables(variables)
        for v in variables:
            v_id = id(v)
            if v_id not in self._referenced_variables:
                raise ValueError(
                    'cannot remove variable {name} - it has not been added'.format(
                        name=v.name
                    )
                )
            cons_using, sos_using, obj_using = self._referenced_variables[v_id]
            if cons_using or sos_using or (obj_using is not None):
                raise ValueError(
                    'cannot remove variable {name} - it is still being used by constraints or the objective'.format(
                        name=v.name
                    )
                )
            del self._referenced_variables[v_id]
            del self._vars[v_id]

    @abc.abstractmethod
    def _remove_params(self, params: List[ParamData]):
        pass

    def remove_params(self, params: List[ParamData]):
        self._remove_params(params)
        for p in params:
            del self._params[id(p)]

    def remove_block(self, block):
        self.remove_constraints(
            [
                con
                for con in block.component_data_objects(
                    ctype=Constraint, descend_into=True, active=True
                )
            ]
        )
        self.remove_sos_constraints(
            [
                con
                for con in block.component_data_objects(
                    ctype=SOSConstraint, descend_into=True, active=True
                )
            ]
        )
        if self._only_child_vars:
            self.remove_variables(
                list(
                    dict(
                        (id(var), var)
                        for var in block.component_data_objects(
                            ctype=Var, descend_into=True
                        )
                    ).values()
                )
            )
        self.remove_params(
            list(
                dict(
                    (id(p), p)
                    for p in block.component_data_objects(
                        ctype=Param, descend_into=True
                    )
                ).values()
            )
        )

    @abc.abstractmethod
    def _update_variables(self, variables: List[VarData]):
        pass

    def update_variables(self, variables: List[VarData]):
        for v in variables:
            self._vars[id(v)] = (
                v,
                v._lb,
                v._ub,
                v.fixed,
                v.domain.get_interval(),
                v.value,
            )
        self._update_variables(variables)

    @abc.abstractmethod
    def update_params(self):
        pass

    def update(self, timer: HierarchicalTimer = None):
        if timer is None:
            timer = HierarchicalTimer()
        config = self.update_config
        new_vars = list()
        old_vars = list()
        new_params = list()
        old_params = list()
        new_cons = list()
        old_cons = list()
        old_sos = list()
        new_sos = list()
        current_vars_dict = dict()
        current_cons_dict = dict()
        current_sos_dict = dict()
        timer.start('vars')
        if self._only_child_vars and (
            config.check_for_new_or_removed_vars or config.update_vars
        ):
            current_vars_dict = {
                id(v): v
                for v in self._model.component_data_objects(Var, descend_into=True)
            }
            for v_id, v in current_vars_dict.items():
                if v_id not in self._vars:
                    new_vars.append(v)
            for v_id, v_tuple in self._vars.items():
                if v_id not in current_vars_dict:
                    old_vars.append(v_tuple[0])
        elif config.update_vars:
            start_vars = {v_id: v_tuple[0] for v_id, v_tuple in self._vars.items()}
        timer.stop('vars')
        timer.start('params')
        if config.check_for_new_or_removed_params:
            current_params_dict = dict()
            for p in self._model.component_objects(Param, descend_into=True):
                if p.mutable:
                    for _p in p.values():
                        current_params_dict[id(_p)] = _p
            for p_id, p in current_params_dict.items():
                if p_id not in self._params:
                    new_params.append(p)
            for p_id, p in self._params.items():
                if p_id not in current_params_dict:
                    old_params.append(p)
        timer.stop('params')
        timer.start('cons')
        if config.check_for_new_or_removed_constraints or config.update_constraints:
            current_cons_dict = {
                c: None
                for c in self._model.component_data_objects(
                    Constraint, descend_into=True, active=True
                )
            }
            current_sos_dict = {
                c: None
                for c in self._model.component_data_objects(
                    SOSConstraint, descend_into=True, active=True
                )
            }
            for c in current_cons_dict.keys():
                if c not in self._vars_referenced_by_con:
                    new_cons.append(c)
            for c in current_sos_dict.keys():
                if c not in self._vars_referenced_by_con:
                    new_sos.append(c)
            for c in self._vars_referenced_by_con.keys():
                if c not in current_cons_dict and c not in current_sos_dict:
                    if (c.ctype is Constraint) or (
                        c.ctype is None and isinstance(c, ConstraintData)
                    ):
                        old_cons.append(c)
                    else:
                        assert (c.ctype is SOSConstraint) or (
                            c.ctype is None and isinstance(c, SOSConstraintData)
                        )
                        old_sos.append(c)
        self.remove_constraints(old_cons)
        self.remove_sos_constraints(old_sos)
        timer.stop('cons')
        timer.start('params')
        self.remove_params(old_params)

        # sticking this between removal and addition
        # is important so that we don't do unnecessary work
        if config.update_params:
            self.update_params()

        self.add_params(new_params)
        timer.stop('params')
        timer.start('vars')
        self.add_variables(new_vars)
        timer.stop('vars')
        timer.start('cons')
        self.add_constraints(new_cons)
        self.add_sos_constraints(new_sos)
        new_cons_set = set(new_cons)
        new_sos_set = set(new_sos)
        new_vars_set = set(id(v) for v in new_vars)
        cons_to_remove_and_add = dict()
        need_to_set_objective = False
        if config.update_constraints:
            for c in current_cons_dict.keys():
                if c not in new_cons_set and c.expr is not self._active_constraints[c]:
                    cons_to_remove_and_add[c] = None
            sos_to_update = []
            for c in current_sos_dict.keys():
                if c not in new_sos_set:
                    sos_to_update.append(c)
            self.remove_sos_constraints(sos_to_update)
            self.add_sos_constraints(sos_to_update)
        timer.stop('cons')
        timer.start('vars')
        if self._only_child_vars and config.update_vars:
            vars_to_check = list()
            for v_id, v in current_vars_dict.items():
                if v_id not in new_vars_set:
                    vars_to_check.append(v)
        elif config.update_vars:
            end_vars = {v_id: v_tuple[0] for v_id, v_tuple in self._vars.items()}
            vars_to_check = [v for v_id, v in end_vars.items() if v_id in start_vars]
        if config.update_vars:
            vars_to_update = list()
            for v in vars_to_check:
                _v, lb, ub, fixed, domain_interval, value = self._vars[id(v)]
                if (fixed != v.fixed) or (fixed and (value != v.value)):
                    vars_to_update.append(v)
                    if self.update_config.treat_fixed_vars_as_params:
                        for c in self._referenced_variables[id(v)][0]:
                            cons_to_remove_and_add[c] = None
                        if self._referenced_variables[id(v)][2] is not None:
                            need_to_set_objective = True
                elif lb is not v._lb:
                    vars_to_update.append(v)
                elif ub is not v._ub:
                    vars_to_update.append(v)
                elif domain_interval != v.domain.get_interval():
                    vars_to_update.append(v)
            self.update_variables(vars_to_update)
        timer.stop('vars')
        timer.start('cons')
        cons_to_remove_and_add = list(cons_to_remove_and_add.keys())
        self.remove_constraints(cons_to_remove_and_add)
        self.add_constraints(cons_to_remove_and_add)
        timer.stop('cons')
        timer.start('named expressions')
        if config.update_named_expressions:
            cons_to_update = list()
            for c, expr_list in self._named_expressions.items():
                if c in new_cons_set:
                    continue
                for named_expr, old_expr in expr_list:
                    if named_expr.expr is not old_expr:
                        cons_to_update.append(c)
                        break
            self.remove_constraints(cons_to_update)
            self.add_constraints(cons_to_update)
            for named_expr, old_expr in self._obj_named_expressions:
                if named_expr.expr is not old_expr:
                    need_to_set_objective = True
                    break
        timer.stop('named expressions')
        timer.start('objective')
        if self.update_config.check_for_new_objective:
            pyomo_obj = get_objective(self._model)
            if pyomo_obj is not self._objective:
                need_to_set_objective = True
        else:
            pyomo_obj = self._objective
        if self.update_config.update_objective:
            if pyomo_obj is not None and pyomo_obj.expr is not self._objective_expr:
                need_to_set_objective = True
            elif pyomo_obj is not None and pyomo_obj.sense is not self._objective_sense:
                # we can definitely do something faster here than resetting the whole objective
                need_to_set_objective = True
        if need_to_set_objective:
            self.set_objective(pyomo_obj)
        timer.stop('objective')

        # this has to be done after the objective and constraints in case the
        # old objective/constraints use old variables
        timer.start('vars')
        self.remove_variables(old_vars)
        timer.stop('vars')


legacy_termination_condition_map = {
    TerminationCondition.unknown: LegacyTerminationCondition.unknown,
    TerminationCondition.maxTimeLimit: LegacyTerminationCondition.maxTimeLimit,
    TerminationCondition.maxIterations: LegacyTerminationCondition.maxIterations,
    TerminationCondition.objectiveLimit: LegacyTerminationCondition.minFunctionValue,
    TerminationCondition.minStepLength: LegacyTerminationCondition.minStepLength,
    TerminationCondition.optimal: LegacyTerminationCondition.optimal,
    TerminationCondition.unbounded: LegacyTerminationCondition.unbounded,
    TerminationCondition.infeasible: LegacyTerminationCondition.infeasible,
    TerminationCondition.infeasibleOrUnbounded: LegacyTerminationCondition.infeasibleOrUnbounded,
    TerminationCondition.error: LegacyTerminationCondition.error,
    TerminationCondition.interrupted: LegacyTerminationCondition.resourceInterrupt,
    TerminationCondition.licensingProblems: LegacyTerminationCondition.licensingProblems,
}


legacy_solver_status_map = {
    TerminationCondition.unknown: LegacySolverStatus.unknown,
    TerminationCondition.maxTimeLimit: LegacySolverStatus.aborted,
    TerminationCondition.maxIterations: LegacySolverStatus.aborted,
    TerminationCondition.objectiveLimit: LegacySolverStatus.aborted,
    TerminationCondition.minStepLength: LegacySolverStatus.error,
    TerminationCondition.optimal: LegacySolverStatus.ok,
    TerminationCondition.unbounded: LegacySolverStatus.error,
    TerminationCondition.infeasible: LegacySolverStatus.error,
    TerminationCondition.infeasibleOrUnbounded: LegacySolverStatus.error,
    TerminationCondition.error: LegacySolverStatus.error,
    TerminationCondition.interrupted: LegacySolverStatus.aborted,
    TerminationCondition.licensingProblems: LegacySolverStatus.error,
}


legacy_solution_status_map = {
    TerminationCondition.unknown: LegacySolutionStatus.unknown,
    TerminationCondition.maxTimeLimit: LegacySolutionStatus.stoppedByLimit,
    TerminationCondition.maxIterations: LegacySolutionStatus.stoppedByLimit,
    TerminationCondition.objectiveLimit: LegacySolutionStatus.stoppedByLimit,
    TerminationCondition.minStepLength: LegacySolutionStatus.error,
    TerminationCondition.optimal: LegacySolutionStatus.optimal,
    TerminationCondition.unbounded: LegacySolutionStatus.unbounded,
    TerminationCondition.infeasible: LegacySolutionStatus.infeasible,
    TerminationCondition.infeasibleOrUnbounded: LegacySolutionStatus.unsure,
    TerminationCondition.error: LegacySolutionStatus.error,
    TerminationCondition.interrupted: LegacySolutionStatus.error,
    TerminationCondition.licensingProblems: LegacySolutionStatus.error,
}


class LegacySolverInterface:
    def solve(
        self,
        model: BlockData,
        tee: bool = False,
        load_solutions: bool = True,
        logfile: Optional[str] = None,
        solnfile: Optional[str] = None,
        timelimit: Optional[float] = None,
        report_timing: bool = False,
        solver_io: Optional[str] = None,
        suffixes: Optional[Sequence] = None,
        options: Optional[Dict] = None,
        keepfiles: bool = False,
        symbolic_solver_labels: bool = False,
        warmstart: bool = False,
    ):
        original_config = self.config
        self.config = self.config()
        self.config.stream_solver = tee
        self.config.load_solution = load_solutions
        self.config.symbolic_solver_labels = symbolic_solver_labels
        self.config.time_limit = timelimit
        self.config.warmstart = warmstart
        self.config.report_timing = report_timing
        if solver_io is not None:
            raise NotImplementedError('Still working on this')
        if suffixes is not None:
            raise NotImplementedError('Still working on this')
        if logfile is not None:
            raise NotImplementedError('Still working on this')
        if 'keepfiles' in self.config:
            self.config.keepfiles = keepfiles
        if solnfile is not None:
            if 'filename' in self.config:
                filename = os.path.splitext(solnfile)[0]
                self.config.filename = filename
        original_options = self.options
        if options is not None:
            self.options = options

        results: Results = super(LegacySolverInterface, self).solve(model)

        legacy_results = LegacySolverResults()
        legacy_soln = LegacySolution()
        legacy_results.solver.status = legacy_solver_status_map[
            results.termination_condition
        ]
        legacy_results.solver.termination_condition = legacy_termination_condition_map[
            results.termination_condition
        ]
        legacy_soln.status = legacy_solution_status_map[results.termination_condition]
        legacy_results.solver.termination_message = str(results.termination_condition)

        obj = get_objective(model)
        legacy_results.problem.sense = obj.sense

        if obj.sense == minimize:
            legacy_results.problem.lower_bound = results.best_objective_bound
            legacy_results.problem.upper_bound = results.best_feasible_objective
        else:
            legacy_results.problem.upper_bound = results.best_objective_bound
            legacy_results.problem.lower_bound = results.best_feasible_objective
        if (
            results.best_feasible_objective is not None
            and results.best_objective_bound is not None
        ):
            legacy_soln.gap = abs(
                results.best_feasible_objective - results.best_objective_bound
            )
        else:
            legacy_soln.gap = None

        symbol_map = SymbolMap()
        symbol_map.byObject = dict(self.symbol_map.byObject)
        symbol_map.bySymbol = dict(self.symbol_map.bySymbol)
        symbol_map.aliases = dict(self.symbol_map.aliases)
        symbol_map.default_labeler = self.symbol_map.default_labeler
        model.solutions.add_symbol_map(symbol_map)
        legacy_results._smap_id = id(symbol_map)

        delete_legacy_soln = True
        if load_solutions:
            if hasattr(model, 'dual') and model.dual.import_enabled():
                for c, val in results.solution_loader.get_duals().items():
                    model.dual[c] = val
            if hasattr(model, 'slack') and model.slack.import_enabled():
                for c, val in results.solution_loader.get_slacks().items():
                    model.slack[c] = val
            if hasattr(model, 'rc') and model.rc.import_enabled():
                for v, val in results.solution_loader.get_reduced_costs().items():
                    model.rc[v] = val
        elif results.best_feasible_objective is not None:
            delete_legacy_soln = False
            for v, val in results.solution_loader.get_primals().items():
                legacy_soln.variable[symbol_map.getSymbol(v)] = {'Value': val}
            if hasattr(model, 'dual') and model.dual.import_enabled():
                for c, val in results.solution_loader.get_duals().items():
                    legacy_soln.constraint[symbol_map.getSymbol(c)] = {'Dual': val}
            if hasattr(model, 'slack') and model.slack.import_enabled():
                for c, val in results.solution_loader.get_slacks().items():
                    symbol = symbol_map.getSymbol(c)
                    if symbol in legacy_soln.constraint:
                        legacy_soln.constraint[symbol]['Slack'] = val
            if hasattr(model, 'rc') and model.rc.import_enabled():
                for v, val in results.solution_loader.get_reduced_costs().items():
                    legacy_soln.variable['Rc'] = val

        legacy_results.solution.insert(legacy_soln)
        if delete_legacy_soln:
            legacy_results.solution.delete(0)

        self.config = original_config
        self.options = original_options

        return legacy_results

    def available(self, exception_flag=True):
        ans = super(LegacySolverInterface, self).available()
        if exception_flag and not ans:
            raise ApplicationError(f'Solver {self.__class__} is not available ({ans}).')
        return bool(ans)

    def license_is_valid(self) -> bool:
        """Test if the solver license is valid on this system.

        Note that this method is included for compatibility with the
        legacy SolverFactory interface.  Unlicensed or open source
        solvers will return True by definition.  Licensed solvers will
        return True if a valid license is found.

        Returns
        -------
        available: bool
            True if the solver license is valid. Otherwise, False.

        """
        return bool(self.available())

    @property
    def options(self):
        for solver_name in ['gurobi', 'ipopt', 'cplex', 'cbc', 'highs', 'maingo']:
            if hasattr(self, solver_name + '_options'):
                return getattr(self, solver_name + '_options')
        raise NotImplementedError('Could not find the correct options')

    @options.setter
    def options(self, val):
        found = False
        for solver_name in ['gurobi', 'ipopt', 'cplex', 'cbc', 'highs', 'maingo']:
            if hasattr(self, solver_name + '_options'):
                setattr(self, solver_name + '_options', val)
                found = True
        if not found:
            raise NotImplementedError('Could not find the correct options')

    def __enter__(self):
        return self

    def __exit__(self, t, v, traceback):
        pass

    @classmethod
    def api_version(self):
        """
        Return the public API supported by this interface.

        Returns
        -------
        ~pyomo.common.enums.SolverAPIVersion
            A solver API enum object
        """
        return SolverAPIVersion.V1


class SolverFactoryClass(Factory):
    def register(self, name, doc=None):
        def decorator(cls):
            self._cls[name] = cls
            self._doc[name] = doc

            class LegacySolver(LegacySolverInterface, cls):
                pass

            LegacySolverFactory.register('appsi_' + name, doc)(LegacySolver)

            return cls

        return decorator

    if TYPE_CHECKING:
        # NOTE: `Factory.__call__` can return None, but for the common case
        def __call__(self, name, **kwds) -> Solver: ...


SolverFactory = SolverFactoryClass()
